from ftw.upgrade.command.jsonapi import add_requestor_authentication_argument
from ftw.upgrade.command.jsonapi import error_handling
from ftw.upgrade.command.jsonapi import with_api_requestor
from ftw.upgrade.command.terminal import TERMINAL


DOCS = """
{t.bold}DESCRIPTION:{t.normal}
    Test authentication, printing the user you are authenticated with.

{t.bold}EXAMPLES:{t.normal}
[quote]
    $ bin/upgrade user --auth admin:admin
    Authenticated as "admin".
[/quote]
""".format(t=TERMINAL).strip()


def setup_argparser(commands):
    command = commands.add_parser(
        'user',
        help='Test authentication and print user.',
        description=DOCS)
    command.set_defaults(func=sites_command)
    add_requestor_authentication_argument(command)


@with_api_requestor
@error_handling
def sites_command(args, requestor):
    response = requestor.GET('current_user')
    print 'Authenticated as "{0}".'.format(response.json())



from AccessControl.SecurityInfo import ClassSecurityInformation
from time import time
import logging


class ProgressLogger(object):
    """Loggs the proggress of a process to the passed
    logger.
    """

    security = ClassSecurityInformation()

    def __init__(self, message, iterable, logger=None,
                 timeout=5):
        self.logger = logger or logging.getLogger('ftw.upgrade')
        self.message = message
        self.iterable = iterable

        if isinstance(iterable, (int, long, float)):
            self.length = iterable
        else:
            self.length = len(iterable)

        self.timeout = timeout
        self._timestamp = None
        self._counter = 0
        self._current_item = None

    security.declarePrivate('__enter__')
    def __enter__(self):
        self.logger.info('STARTING %s' % self.message)
        return self

    security.declarePrivate('__exit__')
    def __exit__(self, exc_type, exc_value, traceback):
        if not exc_type:
            self.logger.info('DONE %s' % self.message)

        else:
            if self._current_item is not None:
                current_step = repr(self._current_item)
            else:
                current_step = 'item nr. %d' % self._counter

            self.logger.error('FAILED %s (%s: %s) at %s' % (
                    self.message,
                    str(exc_type.__name__),
                    str(exc_value),
                    current_step))

    security.declarePrivate('__call__')
    def __call__(self):
        self._counter += 1
        if not self.should_be_logged():
            return

        percent = int(self._counter * 100.0 / self.length)
        self.logger.info('%s of %s (%s%%): %s' % (
                self._counter,
                self.length,
                percent,
                self.message))

    security.declarePrivate('__iter__')
    def __iter__(self):
        with self as step:
            for item in self.iterable:
                self._current_item = item
                yield item
                step()

    security.declarePrivate('should_be_logged')
    def should_be_logged(self):
        now = float(time())

        if self._timestamp is None:
            self._timestamp = now
            return True

        next_stamp = self._timestamp + self.timeout
        if next_stamp <= now:
            self._timestamp = now
            return True

        else:
            return False



from ftw.builder import Builder
from ftw.builder import create
from ftw.upgrade.placefulworkflow import PlacefulWorkflowPolicyActivator
from ftw.upgrade.tests.base import WorkflowTestCase


class TestPlacefulWorkflowPolicyActivator(WorkflowTestCase):

    def setUp(self):
        super(TestPlacefulWorkflowPolicyActivator, self).setUp()

        self.set_workflow_chain(for_type='Folder',
                                to_workflow='intranet_workflow')
        self.set_workflow_chain(for_type='Document',
                                to_workflow='intranet_workflow')

    def test_activate_placeful_workflow_policy_with_mapping(self):
        container = create(Builder('folder')
                           .titled(u'Container')
                           .in_state('external'))

        subfolder = create(Builder('folder')
                           .within(container)
                           .titled(u'Subfolder')
                           .in_state('pending'))

        document = create(Builder('document')
                          .within(subfolder)
                          .titled(u'The Document')
                          .in_state('internally_published'))

        self.create_placeful_workflow_policy(
            named='local_workflow',
            with_workflows={
                'Folder': 'plone_workflow'})

        activator = PlacefulWorkflowPolicyActivator(container)
        activator.activate_policy(
            'local_workflow',
            review_state_mapping={
                ('intranet_workflow', 'plone_workflow'): {
                    'external': 'published',
                    'pending': 'pending'}})

        self.assertReviewStates({
                container: 'published',
                subfolder: 'pending',
                document: 'internally_published'})

    def test_object_security_is_updated(self):
        container = create(Builder('folder')
                           .titled(u'Container')
                           .in_state('external'))

        self.create_placeful_workflow_policy(
            named='local_workflow',
            with_workflows={
                'Folder': 'plone_workflow'})

        activator = PlacefulWorkflowPolicyActivator(container)
        activator.activate_policy(
            'local_workflow',
            review_state_mapping={
                ('intranet_workflow', 'plone_workflow'): {
                    'external': 'published'}})

        self.assertSecurityIsUpToDate()

    def test_object_security_is_reindexed(self):
        container = create(Builder('folder')
                           .titled(u'Container')
                           .in_state('internal'))

        self.assertIn('Member',
                      self.get_allowed_roles_and_users(for_object=container))

        self.create_placeful_workflow_policy(
            named='local_workflow',
            with_workflows={
                'Folder': 'plone_workflow'})

        activator = PlacefulWorkflowPolicyActivator(container)
        activator.activate_policy(
            'local_workflow',
            review_state_mapping={
                ('intranet_workflow', 'plone_workflow'): {
                    'internal': 'published'}})

        self.assertEquals(
            ['Anonymous'],
            self.get_allowed_roles_and_users(for_object=container))



from collections import defaultdict
from contextlib import contextmanager
from copy import deepcopy
from ftw.upgrade.exceptions import CyclicDependencies
from path import Path
from zope.component.hooks import getSite
import logging
import math
import os
import re
import stat
import tarjan.tc
import transaction


def topological_sort(items, partial_order):
    """Perform topological sort.
    items is a list of items to be sorted.
    partial_order is a list of pairs. If pair (a,b) is in it, it means
    that item a should appear before item b.
    Returns a list of the items in one of the possible orders, or None
    if partial_order contains a loop.
    Source: http://www.bitformation.com/art/python_toposort.html
    """

    def add_node(graph, node):
        """Add a node to the graph if not already exists."""
        if node not in graph:
            graph[node] = [0]  # 0 = number of arcs coming into this node.

    def add_arc(graph, fromnode, tonode):
        """Add an arc to a graph. Can create multiple arcs.
        The end nodes must already exist."""
        graph[fromnode].append(tonode)
        # Update the count of incoming arcs in tonode.
        graph[tonode][0] = graph[tonode][0] + 1

    # step 1 - create a directed graph with an arc a->b for each input
    # pair (a,b).
    # The graph is represented by a dictionary. The dictionary contains
    # a pair item:list for each node in the graph. /item/ is the value
    # of the node. /list/'s 1st item is the count of incoming arcs, and
    # the rest are the destinations of the outgoing arcs. For example:
    #           {'a':[0,'b','c'], 'b':[1], 'c':[1]}
    # represents the graph:   c <-- a --> b
    # The graph may contain loops and multiple arcs.
    # Note that our representation does not contain reference loops to
    # cause GC problems even when the represented graph contains loops,
    # because we keep the node names rather than references to the nodes.
    graph = {}
    for v in items:
        add_node(graph, v)
    for a, b in partial_order:
        add_arc(graph, a, b)

    # Step 2 - find all roots (nodes with zero incoming arcs).
    roots = [node for (node, nodeinfo) in graph.items() if nodeinfo[0] == 0]
    roots.sort()

    # step 3 - repeatedly emit a root and remove it from the graph. Removing
    # a node may convert some of the node's direct children into roots.
    # Whenever that happens, we append the new roots to the list of
    # current roots.
    sorted_ = []
    while len(roots) != 0:
        # If len(roots) is always 1 when we get here, it means that
        # the input describes a complete ordering and there is only
        # one possible output.
        # When len(roots) > 1, we can choose any root to send to the
        # output; this freedom represents the multiple complete orderings
        # that satisfy the input restrictions. We arbitrarily take one of
        # the roots using pop(). Note that for the algorithm to be efficient,
        # this operation must be done in O(1) time.
        root = roots.pop()
        sorted_.append(root)
        for child in graph[root][1:]:
            graph[child][0] = graph[child][0] - 1
            if graph[child][0] == 0:
                roots.append(child)
        del graph[root]
    if len(graph.items()) != 0:
        # There is a loop in the input.
        return None
    return sorted_


def find_cyclic_dependencies(dependencies):
    deps = defaultdict(list)
    for first, second in dependencies:
        deps[first].append(second)

    cyclic_dependencies = []
    for name, closure in tarjan.tc.tc(deps).items():
        if name in closure and closure not in cyclic_dependencies:
            cyclic_dependencies.append(closure)

    return cyclic_dependencies


class SizedGenerator(object):

    def __init__(self, generator, length):
        self._length = length
        self._generator = generator

    def __iter__(self):
        return self._generator.__iter__()

    def __len__(self):
        return self._length


class SavepointIterator(object):
    """An iterator that creates a savepoint every n items.

    The goal of this iterator is to move data from the current transaction to
    the disk in order to free up RAM.
    """

    def __init__(self, iterable, threshold, logger=None):
        self.iterable = iterable
        self.threshold = threshold
        self.logger = logger

        if self.logger is None:
            self.logger = logging.getLogger('ftw.upgrade')

        if not threshold:
            raise ValueError("Threshold must be a non-zero value")

    def __iter__(self):
        for i, item in enumerate(self.iterable):
            if i % self.threshold == 0:
                optimize_memory_usage()
                self.logger.info("Created savepoint at %s items" % i)
            yield item

    def __len__(self):
        return self.iterable.__len__()

    @classmethod
    def build(cls, iterable, threshold=None, logger=None):
        if threshold is None:
            threshold = cls.get_default_threshold()

        if threshold:
            return SavepointIterator(iterable, threshold, logger)
        else:
            return iterable

    @staticmethod
    def get_default_threshold():
        """Returns the default savepoint threshold.

        The savepoint iterator threshold can be configured with an environment
        variable ``UPGRADE_SAVEPOINT_THRESHOLD``.
        When set to ``"None"``, savepoints are disabled.
        """
        value = os.environ.get('UPGRADE_SAVEPOINT_THRESHOLD', None)
        if value is None:
            # default unchanged; use application default
            return 1000

        value = value.strip().lower()
        if value == 'none':
            # threshold disabled
            return None

        try:
            value = int(value)
        except ValueError:
            raise ValueError('Invalid savepoint threshold {!r}'.format(value))

        if value > 0:
            return value
        else:
            raise ValueError('Invalid savepoint threshold {!r}'.format(value))


def optimize_memory_usage():
    """Optimizes the current memory usage by garbage collecting objects.

    The function creates a transaction savepoint in order to move pending
    changes from the memory to the disk.
    Afterwards the ZODB connection's pickle cache is notified to garbage
    collect objects, when necessary, so that the configured ZODB cache sizes
    are respected.

    The tradeoff of this function is a low memory footprint versus a speed:
    objects may be removed which will later be used again and have to be loaded
    again from the ZODB.
    """
    transaction.savepoint(optimistic=True)
    # By calling `cacheGC` on the connection, the pickle cache gets a
    # chance to respect the configured zodb cache size by garbage
    # collecting "older" objects (LRU).
    # This only works well when we've created a savepoint in advance,
    # which moves the changes to the disk.
    getSite()._p_jar.cacheGC()


def get_sorted_profile_ids(portal_setup):
    """Returns a sorted list of profile ids (without profile- prefix).
    The sorting is done by resolving the dependencies and performing
    a topological graph sort.
    If there are circular dependencies a CyclicDependencies exception
    is thrown.
    """
    profile_ids = []
    dependencies = []

    for profile in portal_setup.listProfileInfo():
        profile_ids.append(profile['id'])

    for profile in portal_setup.listProfileInfo():
        for dependency in (list(profile.get('dependencies') or []) +
                           list(profile.get('ftw.upgrade:dependencies') or [])):
            dependency = re.sub('^profile-', '', dependency)
            if dependency in profile_ids:
                dependencies.append((profile['id'], dependency))

    order = topological_sort(profile_ids, dependencies)

    if order is None:
        raise CyclicDependencies(
            dependencies,
            find_cyclic_dependencies(deepcopy(dependencies)))
    else:
        return list(reversed(order))


def format_duration(seconds):
    """Makes a duration in seconds human readable.
    Supports hours, minutes and seconds.
    """

    seconds = math.ceil(seconds)
    hours, remainder = divmod(seconds, 60 * 60)
    minutes, seconds = divmod(remainder, 60)

    result = []

    if hours == 1:
        result.append('1 hour')
    elif hours > 1:
        result.append('%i hours' % hours)

    if minutes == 1:
        result.append('1 minute')
    elif minutes > 1:
        result.append('%i minutes' % minutes)

    if seconds == 1:
        result.append('1 second')
    elif seconds > 1:
        result.append('%i seconds' % seconds)

    if len(result) == 0:
        return '0 seconds'
    else:
        return ', '.join(result)


def subject_from_docstring(docstring):
    """Extracts and returns the subject of a docstring.
    The subject consists of all lines from the beginning to the
    first empty line. Newlines are stripped.
    """
    lines = map(str.strip, docstring.strip().splitlines())
    try:
        lines.index('')
    except ValueError:
        pass  # '' is not in list
    else:
        lines = lines[:lines.index('')]

    return ' '.join(lines).strip()


def get_tempfile_authentication_directory(directory=None):
    """Finds the buildout directory and returns the absolute path to the
    relative directory var/ftw.upgrade-authentication/.
    If the directory does not exist it is created.
    """
    directory = Path(directory) or Path.getcwd()
    if not directory.joinpath('bin', 'buildout').isfile():
        return get_tempfile_authentication_directory(directory.parent)

    auth_directory = directory.joinpath('var', 'ftw.upgrade-authentication')
    if not auth_directory.isdir():
        auth_directory.mkdir(mode=0770)

    # Verify that "others" do not have any permissions on this directory.
    if auth_directory.stat().st_mode & stat.S_IRWXO:
        raise ValueError('{0} has invalid mode: "others" should not have '
                         'any permissions'.format(auth_directory))

    return auth_directory


class StartsWithLogFilter(logging.Filter):
    """Filter messages that start with criteria."""

    def __init__(self, criteria):
        self.criteria = criteria

    def filter(self, record):
        return not record.getMessage().startswith(self.criteria)


@contextmanager
def log_silencer(logger_name, criteria):
    """Prevents messages that start with `criteria` from being logged to the
    logger that is registered as `logger_name`.
    """
    log = logging.getLogger(logger_name)
    filt = StartsWithLogFilter(criteria)
    log.addFilter(filt)

    try:
        yield
    finally:
        log.removeFilter(filt)



import pkg_resources

HAS_INDEXING = False

try:
    # Plone 5
    from Products.CMFCore.indexing import processQueue
    from Products.CMFCore.indexing import getQueue
except ImportError:
    try:
        # Plone 4 with collective.indexing
        pkg_resources.get_distribution('collective.indexing')
    except pkg_resources.DistributionNotFound:
        def processQueue():
            # Plone 4 without collective.indexing
            pass
    else:
        from collective.indexing.queue import getQueue
        from collective.indexing.queue import processQueue
        HAS_INDEXING = True
else:
    HAS_INDEXING = True


if HAS_INDEXING:
    from ftw.upgrade.interfaces import IDuringUpgrade
    from ftw.upgrade.progresslogger import ProgressLogger
    from zope.globalrequest import getRequest

    class LoggingQueueProcessor(object):
        """Queue processor to log collective.indexing progress.

        A queue processor is used whenever a collective.indexing queue is
        processed, i.e. when collective.indexing indexes, reindexes or
        unindexes objects in the queue. This may happen several times while
        executing upgrades (e.g. every time when executing a catalog-query).

        For larger deployments with a lot of objects that process may take a
        while, thus we display a progress bar while reindexing.
        """
        should_log = False

        def begin(self):
            self.should_log = IDuringUpgrade.providedBy(getRequest())
            if not self.should_log:
                return

            indexing_queue_length = getQueue().length()
            self.logger = ProgressLogger(
                'Processing indexing queue',
                indexing_queue_length)

        def commit(self):
            pass

        def abort(self):
            pass

        def index(self, obj, attributes):
            if not self.should_log:
                return
            self.logger()

        def reindex(self, obj, attributes, update_metadata=False):
            if not self.should_log:
                return
            self.logger()

        def unindex(self, obj):
            if not self.should_log:
                return
            self.logger()



from datetime import datetime
from datetime import timedelta
from ftw.upgrade.command.terminal import TERMINAL
from ftw.upgrade.directory.scaffold import DATETIME_FORMAT
from ftw.upgrade.directory.scanner import UPGRADESTEP_DATETIME_REGEX
from path import Path
import argparse
import re
import sys


DOCS = """
{t.bold}DESCRIPTION:{t.normal}
    Update the order of upgrades on the filesystem by changing the \
timestamps of the upgrades. \
This does only work with new-style upgrades created with the "create" \
command and used with the "upgrade-step:directory" \
directive.

{t.bold}MOVE TO END:{t.normal}
    By just touching an upgrade it moves to the end of the chain and \
receives the current time as timestamp:
[quote]
$ bin/upgrade touch src/my/package/upgrades/20110101000000_foo
[/quote]

{t.bold}MOVE UPRADE BETWEEN OTHER UPGRADES:{t.normal}
    An upgrade can be moved to a specific position other upgrades by \
using the "--before" or "--after" optional arguments:
[quote]
$ bin/upgrade touch ...upgrades/20110101000000_foo --after \
...upgrades/20120202000000_bar
$ bin/upgrade touch ...upgrades/20110101000000_foo --before \
...upgrades/20120202000000_bar
[/quote]


""".format(t=TERMINAL).strip()


NAME_RE = re.compile(r'^.*/?\d{14}_([^/]*)$')
TIMESTAMP_RE = re.compile(r'^.*/?(\d{14})_[^/]*$')


def setup_argparser(commands):
    command = commands.add_parser(
        'touch',
        help='Update the timestamp of an existing upgrade step.',
        description=DOCS)
    command.set_defaults(func=touch_command)

    command.add_argument('path',
                         help='Path to the upgrade step directory to touch.',
                         type=upgrade_step_path)

    order = command.add_mutually_exclusive_group()
    order.add_argument('--after', '-a',
                       metavar='path',
                       dest='after_path',
                       help='Move after this upgrade step directory.',
                       type=upgrade_step_path)

    order.add_argument('--before', '-b',
                       metavar='path',
                       dest='before_path',
                       help='Move before this upgrade step directory.',
                       type=upgrade_step_path)


def touch_command(args):

    parents = set(map(Path.dirname,
                      filter(bool, (args.path,
                                    args.after_path,
                                    args.before_path))))
    if len(parents) > 1:
        print >>sys.stderr, 'ERORR: All paths must be in the same directory,', \
            'got:'
        for parent in parents:
            print '-', parent
        sys.exit(1)

    new_date = find_new_date(args)
    new_name = NAME_RE.sub(
        r'{0}_\1'.format(new_date.strftime(DATETIME_FORMAT)),
        args.path.name)
    new_path = args.path.dirname().joinpath(new_name)
    args.path.rename(new_path)
    print 'New path:', new_path


def upgrade_step_path(path):
    path = Path(path).abspath()

    if not path.isdir():
        raise argparse.ArgumentTypeError(
            '"{0}" does not exist or is not a directory'.format(path))

    if not path_to_datetime(path):
        raise argparse.ArgumentTypeError(
            '"{0}" has not a valid upgrade step name or does'
            ' not contain an upgrade.py.'.format(path.name))

    return path


def find_new_date(args):
    before = path_to_datetime(args.before_path)
    after = path_to_datetime(args.after_path)
    if not before and not after:
        return datetime.now()

    upgrades = sorted(
        filter(None,
               map(path_to_datetime,
                   Path(args.path.dirname()).glob('*/upgrade.py'))))
    upgrades.remove(path_to_datetime(args.path))

    if after and upgrades[-1] == after:
        return after + timedelta(days=1)

    if before and upgrades[0] == before:
        return before - timedelta(days=1)

    if not before:
        before = upgrades[upgrades.index(after) + 1]

    if not after:
        after = upgrades[upgrades.index(before) - 1]

    return after + (before - after) / 2


def path_to_datetime(path):
    if path is None:
        return None

    if not path.name == 'upgrade.py':
        path = path.joinpath('upgrade.py')

    match = UPGRADESTEP_DATETIME_REGEX.match(path)
    if not match:
        return None

    return datetime.strptime(match.group(1), '%Y%m%d%H%M%S')



from ftw.builder import Builder
from ftw.builder import create
from ftw.upgrade.placefulworkflow import PlacefulWorkflowPolicyActivator
from ftw.upgrade.tests.base import WorkflowTestCase
from ftw.upgrade.workflow import WorkflowChainUpdater
from ftw.upgrade.workflow import WorkflowSecurityUpdater
from Products.CMFCore.utils import getToolByName


class TestWorkflowChainUpdater(WorkflowTestCase):

    def test_changing_workflow_with_mapping(self):
        self.set_workflow_chain(for_type='Folder',
                                to_workflow='intranet_folder_workflow')

        private_folder = create(Builder('folder')
                                .titled(u'Private Folder')
                                .in_state('private'))
        published_folder = create(Builder('folder')
                                  .titled(u'Published Folder')
                                  .in_state('internal'))

        objects = [private_folder, published_folder]
        mapping = {
            ('intranet_folder_workflow', 'folder_workflow'): {
                'private': 'visible',
                'internal': 'published'}}

        with WorkflowChainUpdater(objects, mapping):
            self.set_workflow_chain(for_type='Folder',
                                    to_workflow='folder_workflow')

        self.assertReviewStates({
            private_folder: 'visible',
            published_folder: 'published'})

    def test_object_security_is_updated(self):
        self.set_workflow_chain(for_type='Folder',
                                to_workflow='intranet_workflow')

        container = create(Builder('folder')
                           .titled(u'Container')
                           .in_state('external'))

        mapping = {
            ('intranet_workflow', 'plone_workflow'): {
                'external': 'published'}}

        with WorkflowChainUpdater([container], mapping):
            self.set_workflow_chain(for_type='Folder',
                                    to_workflow='plone_workflow')

        self.assertSecurityIsUpToDate()

    def test_object_security_is_reindexed(self):
        self.set_workflow_chain(for_type='Folder',
                                to_workflow='intranet_workflow')

        container = create(Builder('folder')
                           .titled(u'Container')
                           .in_state('internal'))

        self.assertIn('Member',
                      self.get_allowed_roles_and_users(for_object=container))

        mapping = {
            ('intranet_workflow', 'plone_workflow'): {
                'internal': 'published'}}

        with WorkflowChainUpdater([container], mapping):
            self.set_workflow_chain(for_type='Folder',
                                    to_workflow='plone_workflow')

        self.assertEquals(
            ['Anonymous'],
            self.get_allowed_roles_and_users(for_object=container))

    def test_workflow_history_is_migrated(self):
        self.set_workflow_chain(for_type='Folder',
                                to_workflow='simple_publication_workflow')

        folder = create(Builder('folder'))
        self.execute_transition(folder, 'publish')

        self.assert_workflow_history_length(folder, 2)
        self.assert_workflow_history_entry(folder, {
                'action': 'publish',
                'actor': 'test_user_1_',
                'comments': '',
                'review_state': 'published'})

        mapping = {
            ('simple_publication_workflow', 'intranet_folder_workflow'): {
                'published': 'internal'}}
        with WorkflowChainUpdater([folder], mapping):
            self.set_workflow_chain(for_type='Folder',
                                    to_workflow='intranet_folder_workflow')

        self.assert_workflow_history_length(folder, 2)
        self.assert_workflow_history_entry(folder, {
                'action': 'publish',
                'actor': 'test_user_1_',
                'comments': '',
                'review_state': 'internal'}, -1)

    def test_workflow_history_is_not_migrated_when_migration_disabled(self):
        self.set_workflow_chain(for_type='Folder',
                                to_workflow='simple_publication_workflow')

        folder = create(Builder('folder'))
        self.assert_workflow_history_length(folder, 1)

        mapping = {
            ('simple_publication_workflow', 'folder_workflow'): {
                'published': 'published'}}
        with WorkflowChainUpdater([folder], mapping,
                                  migrate_workflow_history=False):
            self.set_workflow_chain(for_type='Folder',
                                    to_workflow='folder_workflow')

        self.assert_workflow_history_length(folder, 0)

    def test_state_is_set_correctly_when_wfhistory_migration_is_disabled(self):
        self.set_workflow_chain(for_type='Folder',
                                to_workflow='simple_publication_workflow')

        folder = create(Builder('folder'))
        self.assert_workflow_history_length(folder, 1)

        mapping = {
            ('simple_publication_workflow', 'intranet_folder_workflow'): {
                'published': 'internal'}}
        with WorkflowChainUpdater([folder], mapping,
                                  migrate_workflow_history=False):
            self.set_workflow_chain(for_type='Folder',
                                    to_workflow='intranet_folder_workflow')

        self.assert_workflow_history_length(folder, 0)
        self.assertReviewStates({folder: 'internal'})

    def test_workflow_history_transition_is_updated_when_mapped(self):
        self.set_workflow_chain(for_type='Folder',
                                to_workflow='simple_publication_workflow')

        folder = create(Builder('folder'))
        self.execute_transition(folder, 'publish')

        self.assert_workflow_history_length(folder, 2)
        self.assert_workflow_history_entry(folder, {
                'action': 'publish',
                'actor': 'test_user_1_',
                'comments': '',
                'review_state': 'published'})

        mapping = {
            ('simple_publication_workflow', 'intranet_folder_workflow'): {
                'published': 'internal'}}
        transition_mapping = {
            ('simple_publication_workflow', 'intranet_folder_workflow'): {
                'publish': 'show_internally'}}

        with WorkflowChainUpdater([folder], mapping,
                                  transition_mapping=transition_mapping):
            self.set_workflow_chain(for_type='Folder',
                                    to_workflow='intranet_folder_workflow')

        self.assert_workflow_history_length(folder, 2)
        self.assert_workflow_history_entry(folder, {
                'action': 'show_internally',
                'actor': 'test_user_1_',
                'comments': '',
                'review_state': 'internal'}, -1)

    def execute_transition(self, context, transition):
        wftool = getToolByName(self.layer['portal'], 'portal_workflow')
        wftool.doActionFor(context, transition)

    def assert_workflow_history_entry(self, context, expected, index=-1):
        wftool = getToolByName(self.layer['portal'], 'portal_workflow')
        history = wftool.getInfoFor(context, 'review_history')
        self.assertDictContainsSubset(
            expected, history[index],
            'Workflow history entry is wrong.\n{0}'.format(history))

    def assert_workflow_history_length(self, context, length):
        wftool = getToolByName(self.layer['portal'], 'portal_workflow')
        history = wftool.getInfoFor(context, 'review_history')
        self.assertEquals(
            length, len(history),
            'Workflow history length is wrong.\n{0}'.format(history))


class TestWorkflowSecurityUpdater(WorkflowTestCase):

    def test_updates_only_objects_with_specified_workflows(self):
        self.set_workflow_chain(for_type='Folder',
                                to_workflow='folder_workflow')
        folder = create(Builder('folder'))
        folder.manage_permission('View', roles=[],
                                 acquire=True)

        self.set_workflow_chain(for_type='Document',
                                to_workflow='simple_publication_workflow')
        document = create(Builder('document'))
        document.manage_permission('View', roles=[],
                                   acquire=True)

        self.assert_permission_acquired('View', folder)
        self.assert_permission_acquired('View', document)

        updater = WorkflowSecurityUpdater()
        updater.update(['folder_workflow'])

        self.assert_permission_not_acquired(
            'View', folder, 'The folder should have been updated but wasnt.')
        self.assert_permission_acquired(
            'View', document,
            'The document should NOT have been updated but it was.')

    def test_updates_disabling_update_security(self):
        self.set_workflow_chain(for_type='Folder',
                                to_workflow='folder_workflow')
        folder = create(Builder('folder'))
        folder.manage_permission('View', roles=['Reader'], acquire=False)
        folder.reindexObjectSecurity()
        self.assertEquals(['Reader'],
                          self.get_allowed_roles_and_users(for_object=folder))

        updater = WorkflowSecurityUpdater()
        updater.update(['folder_workflow'], reindex_security=False)
        self.assertEquals(['Reader'],
                          self.get_allowed_roles_and_users(for_object=folder))

        updater.update(['folder_workflow'], reindex_security=True)
        self.assertEquals(['Anonymous'],
                          self.get_allowed_roles_and_users(for_object=folder))

    def test_respects_placeful_workflows_when_updating(self):
        container = create(Builder('folder'))
        document = create(Builder('document').within(container))

        self.create_placeful_workflow_policy(
            named='local_workflow',
            with_workflows={'Document': 'simple_publication_workflow'})
        activator = PlacefulWorkflowPolicyActivator(container)
        activator.activate_policy(
            'local_workflow',
            review_state_mapping={
                (None, 'simple_publication_workflow'): {
                    None: 'private'}})

        document.manage_permission('View', roles=[],
                                   acquire=True)
        self.assert_permission_acquired('View', document)

        updater = WorkflowSecurityUpdater()
        updater.update(['simple_publication_workflow'])

        self.assert_permission_not_acquired(
            'View', document,
            'The document should have been updated but was not.')



from datetime import datetime
from ftw.builder import Builder
from ftw.builder import create
from ftw.upgrade.tests.base import CommandAndInstanceTestCase
from ftw.upgrade.tests.helpers import no_logging_threads
from persistent.list import PersistentList
from plone.app.testing import setRoles
from plone.app.testing import TEST_USER_ID
import os
import sys
import transaction


class TestInstallCommand(CommandAndInstanceTestCase):

    def setUp(self):
        super(TestInstallCommand, self).setUp()
        self.write_zconf_with_test_instance()

    def test_help(self):
        self.upgrade_script('install --help')

    def test_install_proposed_upgrades(self):
        self.package.with_profile(
            Builder('genericsetup profile')
            .with_upgrade(Builder('ftw upgrade step').to(datetime(2011, 1, 1))
                          .named('The upgrade')))

        with self.package_created():
            self.install_profile('the.package:default', version='1')
            self.clear_recorded_upgrades('the.package:default')

            self.assertFalse(self.is_installed('the.package:default', datetime(2011, 1, 1)))
            exitcode, output = self.upgrade_script('install -s plone --proposed')
            self.assertEquals(0, exitcode)
            transaction.begin()  # sync transaction
            self.assertTrue(self.is_installed('the.package:default', datetime(2011, 1, 1)))
            self.assertIn('Result: SUCCESS', output)

    def test_skip_deferrable_upgrades_if_specified(self):
        self.package.with_profile(
            Builder('genericsetup profile')
            .with_upgrade(Builder('ftw upgrade step')
                          .to(datetime(2011, 1, 1))
                          .as_deferrable()))

        with self.package_created():
            self.install_profile('the.package:default', version='1')
            self.clear_recorded_upgrades('the.package:default')

            self.assertFalse(self.is_installed('the.package:default', datetime(2011, 1, 1)))
            exitcode, output = self.upgrade_script('install -s plone --proposed --skip-deferrable')
            self.assertEquals(0, exitcode)
            transaction.begin()  # sync transaction
            self.assertFalse(self.is_installed('the.package:default', datetime(2011, 1, 1)))
            self.assertIn('Result: SUCCESS', output)

    def test_install_deferrable_upgrades_by_default(self):
        self.package.with_profile(
            Builder('genericsetup profile')
            .with_upgrade(Builder('ftw upgrade step')
                          .to(datetime(2011, 1, 1))
                          .as_deferrable()))

        with self.package_created():
            self.install_profile('the.package:default', version='1')
            self.clear_recorded_upgrades('the.package:default')

            self.assertFalse(self.is_installed('the.package:default', datetime(2011, 1, 1)))
            exitcode, output = self.upgrade_script('install -s plone --proposed')
            self.assertEquals(0, exitcode)
            transaction.begin()  # sync transaction
            self.assertTrue(self.is_installed('the.package:default', datetime(2011, 1, 1)))
            self.assertIn('Result: SUCCESS', output)

    def test_install_failure_raises_exitcode(self):
        def failing_upgrade(setup_context):
            raise KeyError('foo')

        self.package.with_profile(
            Builder('genericsetup profile')
            .with_upgrade(Builder('plone upgrade step')
                          .upgrading('1', to='2')
                          .calling(failing_upgrade)))

        with self.package_created():
            self.install_profile('the.package:default', version='1')
            with no_logging_threads():
                exitcode, output = self.upgrade_script('install -s plone --proposed',
                                                       assert_exitcode=False)
            self.assertEquals(3, exitcode)
            self.assertIn('Result: FAILURE', output)

    def test_umlauts_can_be_printed(self):
        """Regression:
        https://github.com/4teamwork/ftw.upgrade/issues/141

        When the terminal's encoding is ascii (LC_ALL=C), printing unicode
        strings will result in an encoding error.
        """

        def upgrade(setup_context):
            import logging
            logging.getLogger('The Upgrade').error('Uml\xc3\xa4ut.')

        self.package.with_profile(
            Builder('genericsetup profile')
            .with_upgrade(Builder('plone upgrade step')
                          .upgrading('1', to='2')
                          .calling(upgrade)))

        with self.package_created():
            self.install_profile('the.package:default', version='1')

            # The zc.buildout.testing's system implementation does encoding
            # the command result with the default encoding.
            # Since we need to log umlauts in order to proof that it works
            # we need to make the system implementation support utf-8 characters.
            # It's the easiest way to just temporarily change the system encoding.
            system_encoding = sys.getdefaultencoding()
            reload(sys)
            sys.setdefaultencoding('utf-8')
            try:
                exitcode, output = self.upgrade_script('install -s plone --proposed')
            finally:
                sys.setdefaultencoding(system_encoding)

            self.assertEquals(0, exitcode, output)

    def test_install_list_of_upgrades(self):
        self.package.with_profile(
            Builder('genericsetup profile')
            .with_upgrade(Builder('ftw upgrade step').to(datetime(2011, 1, 1))
                          .named('The upgrade')))

        with self.package_created():
            self.install_profile('the.package:default', version='1')
            self.clear_recorded_upgrades('the.package:default')

            self.assertFalse(self.is_installed('the.package:default', datetime(2011, 1, 1)))
            exitcode, output = self.upgrade_script(
                'install -s plone --upgrades 20110101000000@the.package:default')
            self.assertEquals(0, exitcode)
            transaction.begin()  # sync transaction
            self.assertTrue(self.is_installed('the.package:default', datetime(2011, 1, 1)))
            self.assertIn('Result: SUCCESS', output)

    def test_install_proposed_upgrades_of_profile(self):
        self.package.with_profile(
            Builder('genericsetup profile')
            .with_upgrade(Builder('ftw upgrade step').to(datetime(2011, 1, 1)))
            .with_upgrade(Builder('ftw upgrade step').to(datetime(2011, 2, 2))))

        with self.package_created():
            self.install_profile('the.package:default', version='20110101000000')
            self.clear_recorded_upgrades('the.package:default')
            self.record_installed_upgrades('the.package:default', '20110101000000')

            self.assertTrue(self.is_installed('the.package:default', datetime(2011, 1, 1)))
            self.assertFalse(self.is_installed('the.package:default', datetime(2011, 2, 2)))
            exitcode, output = self.upgrade_script(
                'install -s plone --proposed the.package:default')
            self.assertEquals(0, exitcode)
            transaction.begin()  # sync transaction
            self.assertTrue(self.is_installed('the.package:default', datetime(2011, 1, 1)))
            self.assertTrue(self.is_installed('the.package:default', datetime(2011, 2, 2)))
            self.assertIn('Result: SUCCESS', output)

    def test_install_proposed_upgrades_of_profile_fails_for_invalid_profiles(self):
        exitcode, output = self.upgrade_script(
            'install -s plone --proposed the.inexisting.package:default',
            assert_exitcode=False)
        self.assertEquals(1, exitcode)
        self.assertIn('ERROR:', output)

    def test_virtual_host_monster_is_configured_by_environment_variable(self):
        os.environ['UPGRADE_PUBLIC_URL'] = 'https://foo.bar.com/baz'
        self.layer['portal'].upgrade_info = PersistentList()

        setRoles(self.layer['portal'], TEST_USER_ID, ['Manager'])
        create(Builder('folder').titled(u'The Folder'))

        def upgrade_step(setup_context):
            portal = setup_context.portal_url.getPortalObject()
            folder = portal.get('the-folder')
            portal.upgrade_info.append(folder.absolute_url())

        self.package.with_profile(
            Builder('genericsetup profile')
            .with_upgrade(Builder('plone upgrade step').upgrading('1', to='2')
                          .calling(upgrade_step)))

        with self.package_created():
            self.install_profile('the.package:default', version='1')
            exitcode, output = self.upgrade_script('install -s plone --proposed')
            self.assertEquals(0, exitcode)
            transaction.begin()  # sync transaction
            self.assertEquals(['https://foo.bar.com/baz/the-folder'],
                              self.layer['portal'].upgrade_info)

    def test_install_profiles(self):
        self.package.with_profile(Builder('genericsetup profile'))

        self.setup_logging()
        with self.package_created():
            exitcode, output = self.upgrade_script(
                'install -s plone --profiles the.package:default')
            self.assertEquals(
                [u'ftw.upgrade: Installing profile the.package:default.',
                 u'ftw.upgrade: Done installing profile the.package:default.',
                 u'Result: SUCCESS'],
                output.splitlines())

    def test_install_profiles_skipped_when_already_installed(self):
        self.setup_logging()
        self.purge_log()
        exitcode, output = self.upgrade_script(
            'install -s plone --profiles ftw.upgrade:default')
        self.assertEquals(
            [u'ftw.upgrade: Ignoring already installed profile'
             u' ftw.upgrade:default.',
             u'Result: SUCCESS'],
            output.splitlines())

    def test_force_install_already_installed_profiles(self):
        self.setup_logging()
        self.purge_log()
        exitcode, output = self.upgrade_script(
            'install -s plone --force --profiles ftw.upgrade:default')
        self.assertEquals(
            [u'ftw.upgrade: Installing profile ftw.upgrade:default.',
             u'ftw.upgrade: Done installing profile ftw.upgrade:default.',
             u'Result: SUCCESS'],
            output.splitlines())

    def test_force_option_is_meant_to_be_combined_with_profiles(self):
        exitcode, output = self.upgrade_script(
            'install -s plone --force --upgrades 20110101000000@the.package:default',
            assert_exitcode=False)
        self.assertEquals(3, exitcode)
        self.assertEquals(
            [u'ERROR: --force can only be used with --profiles.'],
            output.splitlines())



from ftw.upgrade.directory.scanner import Scanner
from ftw.upgrade.directory.wrapper import wrap_upgrade_step
from ftw.upgrade.exceptions import UpgradeStepConfigurationError
from operator import attrgetter
from Products.CMFPlone.interfaces import IMigratingPloneSiteRoot
from Products.GenericSetup.interfaces import EXTENSION
from Products.GenericSetup.interfaces import IProfile
from Products.GenericSetup.registry import _profile_registry
from Products.GenericSetup.registry import GlobalRegistryStorage
from Products.GenericSetup.upgrade import _registerUpgradeStep
from Products.GenericSetup.upgrade import _upgrade_registry
from Products.GenericSetup.upgrade import UpgradeStep
from zope.configuration.fields import Path
from zope.configuration.fields import Tokens
from zope.interface import Interface
import os
import zope.schema


class IUpgradeStepDirectoryDirective(Interface):

    profile = zope.schema.TextLine(
        title=u"GenericSetup profile id",
        required=True)

    directory = Path(
        title=u'Path to the upgrade steps directory',
        required=True)

    soft_dependencies = Tokens(
        title=u'List of Generic Setup profile dependencies.',
        description=u'Format: "my.package:default"',
        required=False,
        value_type=zope.schema.TextLine())


def upgrade_step_directory_handler(context, profile, directory,
                                   soft_dependencies=None):
    dottedname = context.package.__name__
    package_dir = os.path.dirname(context.package.__file__)
    if package_dir != os.path.abspath(directory):
        dottedname += '.' + '.'.join(
            os.path.relpath(os.path.abspath(directory), package_dir)
            .split(os.sep))

    context.action(
        discriminator=('upgrade-step:directory', profile),
        callable=upgrade_step_directory_action,
        args=(profile, dottedname, context.path(directory),
              soft_dependencies))


def upgrade_step_directory_action(profile, dottedname, path,
                                  soft_dependencies):
    start_version = find_start_version(profile)
    scanner = Scanner(dottedname, path)

    if profile not in _profile_registry.listProfiles():
        raise UpgradeStepConfigurationError(
            'The profile "{0}" needs to be registered before registering its'
            ' upgrade step directory.'.format(profile))

    profileinfo = _profile_registry.getProfileInfo(profile)
    if profileinfo.get('version', None) is not None:
        raise UpgradeStepConfigurationError(
            'Registering an upgrades directory for "{0}" requires this profile'
            ' to not define a version in its metadata.xml.'
            ' The version is automatically set to the latest upgrade.'.format(
                profile))

    _package, profilename = profile.split(':', 1)
    last_version = ''.join(find_start_version(profile))
    for upgrade_info in scanner.scan():
        upgrade_profile_name = '{0}-upgrade-{1}'.format(
            profilename, upgrade_info['target-version'])

        upgrade_handler = wrap_upgrade_step(
            handler=upgrade_info['callable'],
            upgrade_profile='profile-{0}:{1}'.format(dottedname,
                                                     upgrade_profile_name),
            base_profile=profile,
            target_version=upgrade_info['target-version'])

        step = UpgradeStep(upgrade_info['title'],
                           profile,
                           upgrade_info['source-version'] or start_version,
                           upgrade_info['target-version'],
                           '',
                           upgrade_handler)
        _registerUpgradeStep(step)

        _profile_registry.registerProfile(
            name=upgrade_profile_name,
            title='Upgrade {0} to {1}: {2}'.format(
                profile,
                upgrade_info['target-version'],
                upgrade_info['title']),
            description='',
            path=upgrade_info['path'],
            product=dottedname,
            profile_type=EXTENSION,
            for_=IMigratingPloneSiteRoot)

        last_version = upgrade_info['target-version']

    profile = GlobalRegistryStorage(IProfile).get(profile)
    profile['version'] = last_version
    profile['ftw.upgrade:dependencies'] = soft_dependencies


def find_start_version(profile):
    upgrades = _upgrade_registry.getUpgradeStepsForProfile(profile).values()
    upgrades = sorted(upgrades, key=attrgetter('dest'))
    if len(upgrades) > 0:
        return upgrades[-1].dest
    else:
        return str(10 ** 13)



from datetime import datetime
from ftw.builder import Builder
from ftw.upgrade.tests.base import CommandAndInstanceTestCase
import json
import re


class TestListCommand(CommandAndInstanceTestCase):

    def setUp(self):
        super(TestListCommand, self).setUp()
        self.write_zconf_with_test_instance()

    def test_help(self):
        self.upgrade_script('list --help')

    def test_listing_profiles(self):
        self.package.with_profile(
            Builder('genericsetup profile')
            .with_upgrade(Builder('ftw upgrade step').to(datetime(2011, 1, 1)))
            .with_upgrade(Builder('ftw upgrade step').to(datetime(2012, 2, 2))))

        with self.package_created():
            self.install_profile('the.package:default', version='20110101000000')
            self.clear_recorded_upgrades('the.package:default')

            exitcode, output = self.upgrade_script('list --profiles -s plone')
            self.assertEquals(0, exitcode)

            normalized_output = map(unicode.strip,
                                    re.sub(r' +', ' ', output).splitlines())
            self.assertIn('Installed profiles:', normalized_output)
            self.assertIn('the.package:default'
                          ' 2 proposed 1 orphan'
                          ' the.package'
                          ' 20110101000000 / 20120202000000',
                          normalized_output)

    def test_listing_profiles_as_json(self):
        self.package.with_profile(
            Builder('genericsetup profile')
            .with_upgrade(Builder('ftw upgrade step').to(datetime(2011, 1, 1)))
            .with_upgrade(Builder('ftw upgrade step').to(datetime(2012, 2, 2))))

        with self.package_created():
            self.install_profile('the.package:default', version='20110101000000')
            self.clear_recorded_upgrades('the.package:default')

            exitcode, output = self.upgrade_script('list --profiles -s plone --json')
            self.assertEquals(0, exitcode)
            self.assert_json_contains_profile(
                {'id': 'the.package:default',
                 'title': 'the.package',
                 'product': 'the.package',
                 'db_version': '20110101000000',
                 'fs_version': '20120202000000',
                 'outdated_fs_version': False,
                 'upgrades': [
                        {'id': '20110101000000@the.package:default',
                         'title': 'Upgrade.',
                         'source': '10000000000000',
                         'dest': '20110101000000',
                         'proposed': True,
                         'deferrable': False,
                         'done': False,
                         'orphan': True,
                         'outdated_fs_version': False},
                        {'id': '20120202000000@the.package:default',
                         'title': 'Upgrade.',
                         'source': '20110101000000',
                         'dest': '20120202000000',
                         'proposed': True,
                         'deferrable': False,
                         'done': False,
                         'orphan': False,
                         'outdated_fs_version': False},
                        ]},
                json.loads(output))

    def test_listing_proposed_upgrades(self):
        self.package.with_profile(
            Builder('genericsetup profile')
            .with_upgrade(Builder('ftw upgrade step').to(datetime(2011, 1, 1)))
            .with_upgrade(Builder('ftw upgrade step').to(datetime(2012, 2, 2))))

        with self.package_created():
            self.install_profile('the.package:default', version='20110101000000')
            self.clear_recorded_upgrades('the.package:default')

            exitcode, output = self.upgrade_script('list --upgrades -s plone')
            self.assertEquals(0, exitcode)
            self.assertMultiLineEqual(
                'Proposed upgrades:\n'
                'ID:                                        Title:    \n'
                '20110101000000@the.package:default ORPHAN  Upgrade.  \n'
                '20120202000000@the.package:default         Upgrade.  \n',
                output)

    def test_listing_proposed_upgrades_as_json(self):
        self.package.with_profile(
            Builder('genericsetup profile')
            .with_upgrade(Builder('ftw upgrade step').to(datetime(2011, 1, 1)))
            .with_upgrade(Builder('ftw upgrade step').to(datetime(2012, 2, 2))))

        with self.package_created():
            self.install_profile('the.package:default', version='20110101000000')
            self.clear_recorded_upgrades('the.package:default')

            exitcode, output = self.upgrade_script('list --upgrades -s plone --json')
            self.assertEquals(0, exitcode)
            self.assert_json_equal(
                [{
                        "dest": "20110101000000",
                        "done": False,
                        "id": "20110101000000@the.package:default",
                        "orphan": True,
                        "outdated_fs_version": False,
                        "proposed": True,
                        "deferrable": False,
                        "source": "10000000000000",
                        "title": "Upgrade."
                        },
                 {
                        "dest": "20120202000000",
                        "done": False,
                        "id": "20120202000000@the.package:default",
                        "orphan": False,
                        "outdated_fs_version": False,
                        "proposed": True,
                        "deferrable": False,
                        "source": "20110101000000",
                        "title": "Upgrade."
                        }],
                json.loads(output))

    def test_deferrable_upgrades_are_annotated_in_list(self):
        self.package.with_profile(
            Builder('genericsetup profile')
            .with_upgrade(Builder('ftw upgrade step')
                          .to(datetime(2011, 1, 1))
                          .as_deferrable()))

        with self.package_created():
            self.install_profile('the.package:default', version='1')
            self.clear_recorded_upgrades('the.package:default')

            exitcode, output = self.upgrade_script('list --upgrades -s plone')
            self.assertEquals(0, exitcode)

            self.assertMultiLineEqual(
                u'Proposed upgrades:\n'
                'ID:                                            Title:             \n'
                '20110101000000@the.package:default DEFERRABLE  DeferrableUpgrade  \n',
                output)

    def test_orphaned_is_omitted_in_listing_for_deferrable_upgrades(self):
        self.package.with_profile(
            Builder('genericsetup profile')
            .with_upgrade(Builder('ftw upgrade step').to(datetime(2011, 1, 1)).as_deferrable())
            .with_upgrade(Builder('ftw upgrade step').to(datetime(2012, 2, 2))))

        with self.package_created():
            self.install_profile('the.package:default', version='20110101000000')
            self.clear_recorded_upgrades('the.package:default')

            exitcode, output = self.upgrade_script('list --upgrades -s plone')
            self.assertEquals(0, exitcode)
            self.assertMultiLineEqual(
                'Proposed upgrades:\n'
                'ID:                                            Title:             \n'
                '20110101000000@the.package:default DEFERRABLE  DeferrableUpgrade  \n'
                '20120202000000@the.package:default             Upgrade.           \n',
                output)

    def test_listing_deferrable_upgrades_as_json(self):
        self.package.with_profile(
            Builder('genericsetup profile')
            .with_upgrade(Builder('ftw upgrade step')
                          .to(datetime(2011, 1, 1))
                          .as_deferrable()))

        with self.package_created():
            self.install_profile('the.package:default', version='1')
            self.clear_recorded_upgrades('the.package:default')

            exitcode, output = self.upgrade_script('list --upgrades -s plone --json')
            self.assertEquals(0, exitcode)
            self.assert_json_equal(
                [{
                        "dest": "20110101000000",
                        "done": False,
                        "id": "20110101000000@the.package:default",
                        "orphan": False,
                        "outdated_fs_version": False,
                        "proposed": True,
                        "deferrable": True,
                        "source": "10000000000000",
                        "title": "DeferrableUpgrade"
                        }],
                json.loads(output))

    def test_pick_site_argument(self):
        self.package.with_profile(
            Builder('genericsetup profile')
            .with_upgrade(Builder('ftw upgrade step').to(datetime(2011, 1, 1))))

        with self.package_created():
            self.install_profile('the.package:default', version='20110101000000')
            self.clear_recorded_upgrades('the.package:default')

            exitcode, output = self.upgrade_script('list --upgrades --json --pick-site')
            self.assertEqual(1, len(json.loads(output)))

    def test_last_site_argument(self):
        self.package.with_profile(
            Builder('genericsetup profile')
            .with_upgrade(Builder('ftw upgrade step').to(datetime(2011, 1, 1))))

        with self.package_created():
            self.install_profile('the.package:default', version='20110101000000')
            self.clear_recorded_upgrades('the.package:default')

            exitcode, output = self.upgrade_script('list --upgrades --json --pick-site')
            self.assertEqual(1, len(json.loads(output)))

    def test_all_sites_argument_json(self):
        self.package.with_profile(
            Builder('genericsetup profile')
            .with_upgrade(
                Builder('ftw upgrade step').to(datetime(2011, 1, 1))))

        with self.package_created():
            self.install_profile(
                'the.package:default', version='20110101000000')
            self.clear_recorded_upgrades('the.package:default')

            exitcode, output = self.upgrade_script(
                'list --upgrades --json --all-sites')
            self.assertEqual(exitcode, 0)
            decoded = json.loads(output)
            self.assertEqual(len(decoded), 1)
            # The Plone site id is used as key.
            self.assertEqual(decoded.keys(), ['/plone'])
            self.assertTrue(isinstance(decoded['/plone'], list))

    def test_all_sites_argument_no_json(self):
        self.package.with_profile(
            Builder('genericsetup profile')
            .with_upgrade(
                Builder('ftw upgrade step').to(datetime(2011, 1, 1))))

        with self.package_created():
            self.install_profile(
                'the.package:default', version='20110101000000')
            self.clear_recorded_upgrades('the.package:default')

            exitcode, output = self.upgrade_script(
                'list --upgrades --all-sites')
            self.assertEqual(exitcode, 0)
            self.assertIn('Proposed upgrades', output)
            self.assertIn('20110101000000@the.package:default', output)
            self.assertIn('INFO: Acting on site /plone', output)



from StringIO import StringIO
from ftw.upgrade.progresslogger import ProgressLogger
from time import sleep
from unittest2 import TestCase
import logging


class TestProgressLogger(TestCase):

    def setUp(self):
        self.log = StringIO()
        self.logger = logging.getLogger('ftw.upgrade')
        self.logger.setLevel(logging.DEBUG)

        handler = logging.StreamHandler(self.log)
        self.logger.addHandler(handler)

    def read_log(self):
        self.log.seek(0)
        return self.log.read().strip().split('\n')

    def test_succeeding_logging(self):
        with ProgressLogger('Foo', 5, logger=self.logger,
                            timeout=0.03) as step:
            for i in range(5):
                step()
                sleep(0.0151)

        self.assertEqual(['STARTING Foo',
                          '1 of 5 (20%): Foo',
                          '3 of 5 (60%): Foo',
                          '5 of 5 (100%): Foo',
                          'DONE Foo'],
                         self.read_log())

    def test_failing_logging(self):
        timeout = 0

        with self.assertRaises(ValueError):

            data = range(5)

            with ProgressLogger('Bar', data, logger=self.logger,
                                timeout=timeout) as step:
                for i in data:
                    if i == 3:
                        raise ValueError('baz')

                    step()

        self.assertEqual(['STARTING Bar',
                          '1 of 5 (20%): Bar',
                          '2 of 5 (40%): Bar',
                          '3 of 5 (60%): Bar',
                          'FAILED Bar (ValueError: baz) at item nr. 3'],
                         self.read_log())

    def test_accepts_iterable_object(self):
        items = range(5)

        with ProgressLogger('Foo', items, logger=self.logger) as step:
            for _item in items:
                step()

        self.assertEqual(['STARTING Foo',
                          '1 of 5 (20%): Foo',
                          'DONE Foo'],
                         self.read_log())

    def test_acts_as_iterable_wrapper(self):
        items = range(5)

        result = []

        for item in ProgressLogger('Foo', items, logger=self.logger):
            result.append(item)

        self.assertEqual(['STARTING Foo',
                          '1 of 5 (20%): Foo',
                          'DONE Foo'],
                         self.read_log())

        self.assertEqual(
            items, result,
            'Iterating over the progresslogger yields the original items.')

    def test_current_item_is_printed_when_logger_exits_unexpectedly(self):
        items = range(5)

        with self.assertRaises(ValueError):
            for item in ProgressLogger('Foo', items, logger=self.logger):
                if item == 4:
                    raise ValueError('baz')

        self.assertEquals(['STARTING Foo',
                           '1 of 5 (20%): Foo',
                           'FAILED Foo (GeneratorExit: ) at 4'],
                          self.read_log())



from ftw.upgrade.gatherer import flatten_upgrades
from ftw.upgrade.interfaces import IUpgradeStepRecorder
from operator import itemgetter
from Products.CMFCore.utils import getToolByName
from zope.component import getMultiAdapter
import re


def profile_installed(event):
    if not event.profile_id:
        return

    profile = re.sub('^profile-', '', event.profile_id)
    portal = getToolByName(event.tool, 'portal_url').getPortalObject()
    recorder = getMultiAdapter((portal, profile), IUpgradeStepRecorder)

    map(recorder.mark_as_installed,
        map(itemgetter('sdest'),
            flatten_upgrades(event.tool.listUpgrades(profile, show_old=True))))




class APIError(Exception):
    def __init__(self, message, details='', response_code=400):
        super(APIError, self).__init__(message)
        self.message = message
        self.details = details
        self.response_code = response_code

    def process_error(self, response):
        return


class WrongAPIVersion(APIError):
    def __init__(self, requested_version):
        super(WrongAPIVersion, self).__init__(
            'Wrong API version',
            'The API version "{0}" is not available.'.format(
                requested_version),
            response_code=404)


class UnkownAPIAction(APIError):
    def __init__(self, action_name):
        super(UnkownAPIAction, self).__init__(
            'Unkown API action',
            'There is no API action "{0}".'.format(action_name),
            response_code=404)


class UnauthorizedWrapper(APIError):
    def __init__(self, original_exception):
        super(UnauthorizedWrapper, self).__init__(
            'Unauthorized',
            'Admin authorization required.',
            response_code=401)


class MethodNotAllowed(APIError):
    def __init__(self, required_method):
        self.required_method = required_method.upper()
        super(MethodNotAllowed, self).__init__(
            'Method Not Allowed',
            'Action requires {0}'.format(self.required_method),
            response_code=405)

    def process_error(self, response):
        response.setHeader('Allow', self.required_method)


class MissingParam(APIError):
    def __init__(self, param_name):
        super(MissingParam, self).__init__(
            'Param missing',
            'The param "{0}" is required for this API action.'.format(
                param_name))


class PloneSiteOutdated(APIError):
    def __init__(self):
        super(PloneSiteOutdated, self).__init__(
            'Plone site outdated',
            'The Plone site is outdated and needs to be upgraded'
            ' first using the regular Plone upgrading tools.')


class CyclicDependenciesWrapper(APIError):
    def __init__(self, original_exception):
        super(CyclicDependenciesWrapper, self).__init__(
            'Cyclic dependencies',
            'There are cyclic Generic Setup profile dependencies.',
            response_code=500)


class ProfileNotAvailable(APIError):
    def __init__(self, profileid):
        super(ProfileNotAvailable, self).__init__(
            'Profile not available',
            'The profile "{0}" is wrong or not installed'
            ' on this Plone site.'.format(profileid))


class ProfileNotFound(APIError):
    def __init__(self, profileid):
        super(ProfileNotFound, self).__init__(
            'Profile not found',
            'The profile "{0}" is unknown.'.format(profileid))


class UpgradeNotFoundWrapper(APIError):
    def __init__(self, original_exception):
        api_upgrade_id = original_exception.api_id
        super(UpgradeNotFoundWrapper, self).__init__(
            'Upgrade not found',
            'The upgrade "{0}" is unkown.'.format(api_upgrade_id))


class AbortTransactionWithStreamedResponse(Exception):
    """This exception wraps another exception and is used to indicate that
    the original exception should cause the transaction to be aborted but
    not cause 500 since we are streaming a response.
    It is expected that the exception information (e.g. the traceback) is
    already written to the response and streamed to the browser.
    """

    def __init__(self, original_exception):
        self.original_exception = original_exception



from datetime import datetime
from ftw.builder import Builder
from ftw.upgrade import UpgradeStep
from ftw.upgrade.tests.base import UpgradeTestCase
from Products.CMFCore.utils import getToolByName


class TestUpgradeStepBuilder(UpgradeTestCase):

    def setUp(self):
        super(TestUpgradeStepBuilder, self).setUp()
        self.profile = Builder('genericsetup profile')
        self.package.with_profile(self.profile)

    def test_upgrade_step_directory_and_file_is_created(self):
        self.profile.with_upgrade(Builder('ftw upgrade step')
                                  .to(datetime(2011, 1, 1))
                                  .named('migrate file content type'))

        with self.package_created() as package:
            upgrade_path = package.package_path.joinpath(
                'upgrades', '20110101000000_migrate_file_content_type')
            self.assertTrue(upgrade_path.isdir(),
                            'Upgrade directory was not created {0}'.format(upgrade_path))
            self.assertMultiLineEqual(
                '\n'.join(('from ftw.upgrade import UpgradeStep',
                           '',
                           '',
                           'class MigrateFileContentType(UpgradeStep):',
                           '    """Migrate file content type.',
                           '    """',
                           '',
                           '    def __call__(self):',
                           '        self.install_upgrade_profile()',
                           '')),
                upgrade_path.joinpath('upgrade.py').text())

    def test_executing_upgrade_step_with_custom_code(self):
        class AddExcludeFromNavIndex(UpgradeStep):
            def __call__(self):
                self.catalog_add_index('excludeFromNav', 'KeywordIndex')

        self.profile.with_upgrade(Builder('ftw upgrade step')
                                  .to(datetime(2011, 1, 1))
                                  .calling(AddExcludeFromNavIndex))

        catalog = getToolByName(self.portal, 'portal_catalog')
        with self.package_created():
            self.install_profile('the.package:default', '0')
            self.assertNotIn('excludeFromNav', catalog.indexes(),
                             'Index excludeFromNav already exists.')
            self.install_profile_upgrades('the.package:default')
            self.assertIn('excludeFromNav', catalog.indexes(),
                          'Index excludeFromNav was not created.')

    def test_add_files_and_directories_to_profile(self):
        self.profile.with_upgrade(
            Builder('ftw upgrade step')
            .to(datetime(2011, 1, 1))
            .with_file('foo.txt', 'FOO')
            .with_directory('bar')
            .with_file('baz/baz.txt', 'BAZ', makedirs=True))

        with self.package_created() as package:
            upgrade_path = package.package_path.joinpath('upgrades',
                                                         '20110101000000_upgrade')
            self.assertTrue(upgrade_path.isdir(),
                            'Upgrade directory was not created {0}'.format(upgrade_path))

            self.assertEqual('FOO', upgrade_path.joinpath('foo.txt').text())
            self.assertTrue(upgrade_path.joinpath('bar').isdir(),
                            'directory "bar" was not created.')
            self.assertEqual('BAZ', upgrade_path.joinpath('baz', 'baz.txt').text())

    def test_importing_upgrade_step_with_import_profile_files(self):
        self.profile.with_upgrade(
            Builder('ftw upgrade step')
            .to(datetime(2011, 1, 1))
            .with_file('properties.xml', self.asset('foo-property.xml')))

        with self.package_created():
            self.install_profile('the.package:default', '0')
            self.assertFalse(self.portal.getProperty('foo'),
                             'Expected property "foo" to not yet exist.')
            self.install_profile_upgrades('the.package:default')
            self.assertEqual('bar',
                             self.portal.getProperty('foo'),
                             'Property "foo" was not created.')



from ftw.upgrade.browser.manage import ResponseLogger
from ftw.upgrade.interfaces import IExecutioner
from ftw.upgrade.interfaces import IUpgradeInformationGatherer
from ftw.upgrade.jsonapi.base import APIView
from ftw.upgrade.jsonapi.exceptions import AbortTransactionWithStreamedResponse
from ftw.upgrade.jsonapi.exceptions import PloneSiteOutdated
from ftw.upgrade.jsonapi.exceptions import ProfileNotAvailable
from ftw.upgrade.jsonapi.exceptions import ProfileNotFound
from ftw.upgrade.jsonapi.utils import action
from ftw.upgrade.jsonapi.utils import jsonify
from ftw.upgrade.jsonapi.utils import parse_bool
from ftw.upgrade.resource_registries import recook_resources
from operator import itemgetter
from Products.CMFCore.utils import getToolByName


class PloneSiteAPI(APIView):

    def __init__(self, *args, **kwargs):
        super(PloneSiteAPI, self).__init__(*args, **kwargs)
        self.portal_setup = getToolByName(self.context, 'portal_setup')
        self.gatherer = IUpgradeInformationGatherer(self.portal_setup)

    @jsonify
    @action('GET')
    def get_profile(self, profileid):
        """Returns the profile with the id "profileid" as hash.
        """
        return self._refine_profile_info(self._get_profile_info(profileid))

    @jsonify
    @action('GET')
    def list_profiles(self):
        """Returns a list of all installed profiles.
        """
        return map(self._refine_profile_info, self.gatherer.get_profiles())

    @jsonify
    @action('GET')
    def list_profiles_proposing_upgrades(self):
        """Returns a list of profiles with proposed upgrade steps.
        The upgrade steps of each profile only include proposed upgrades.
        """
        return map(self._refine_profile_info,
                   self.gatherer.get_profiles(proposed_only=True))

    @jsonify
    @action('GET')
    def list_proposed_upgrades(self, propose_deferrable=True):
        """Returns a list of proposed upgrades.
        """
        propose_deferrable = parse_bool(propose_deferrable)
        return map(
            self._refine_upgrade_info,
            self._get_proposed_upgrades(propose_deferrable=propose_deferrable))

    @action('POST', rename_params={'upgrades': 'upgrades:list'})
    def execute_upgrades(self, upgrades):
        """Executes a list of upgrades, each identified by the upgrade ID
        in the form "[dest-version]@[profile ID]".
        """
        self._require_up_to_date_plone_site()
        self._validate_upgrade_ids(*upgrades)
        return self._install_upgrades(*upgrades)

    @action('POST', rename_params={'profiles': 'profiles:list'})
    def execute_proposed_upgrades(self, profiles=None, propose_deferrable=True):
        """Executes all proposed upgrades.
        """
        self._require_up_to_date_plone_site()
        if profiles:
            self._validate_profile_ids(*profiles)
        propose_deferrable = parse_bool(propose_deferrable)

        api_ids = map(itemgetter('api_id'), self._get_proposed_upgrades(
            only_profiles=profiles, propose_deferrable=propose_deferrable))
        return self._install_upgrades(
            *api_ids, propose_deferrable=propose_deferrable)

    @action('POST', rename_params={'profiles': 'profiles:list'})
    def execute_profiles(self, profiles, force_reinstall=False):
        """Executes a list of profiles, each identified by their ID.
        """
        self._require_up_to_date_plone_site()
        self._validate_profile_ids(*profiles)
        return self._install_profiles(*profiles,
                                      force_reinstall=force_reinstall)

    @jsonify
    @action('POST')
    def recook_resources(self):
        """Recook CSS and JavaScript resource bundles.
        """
        recook_resources()
        return 'OK'

    @jsonify
    @action('POST')
    def plone_upgrade(self):
        """Upgrade the Plone Site.

        This is what you would manually do in the @@plone-upgrade view.
        """
        portal_migration = getToolByName(self.context, 'portal_migration')
        if not portal_migration.needUpgrading():
            return 'Plone Site was already up to date.'
        portal_migration.upgrade(swallow_errors=False)
        return 'Plone Site has been updated.'

    @jsonify
    @action('GET')
    def plone_upgrade_needed(self):
        """Returns "true" when Plone needs to be upgraded.
        """
        portal_migration = getToolByName(self.context, 'portal_migration')
        return bool(portal_migration.needUpgrading())

    def _refine_profile_info(self, profile):
        return {'id': profile['id'],
                'title': profile['title'],
                'product': profile['product'],
                'db_version': profile['db_version'],
                'fs_version': profile['version'],
                'outdated_fs_version': False,
                'upgrades': map(self._refine_upgrade_info,
                                profile['upgrades'])}

    def _refine_upgrade_info(self, upgrade):
        keys = ('title', 'proposed', 'deferrable', 'done', 'orphan',
                'outdated_fs_version')
        values = dict((key, value) for (key, value) in upgrade.items()
                      if key in keys)
        values.update({'id': upgrade['api_id'],
                       'title': upgrade['title'],
                       'source': upgrade['ssource'],
                       'dest': upgrade['sdest']})
        return values

    def _get_profile_info(self, profileid):
        profiles = filter(lambda profile: profile['id'] == profileid,
                          self.gatherer.get_profiles())
        if len(profiles) == 0:
            raise ProfileNotAvailable(profileid)
        else:
            return profiles[0]

    def _get_proposed_upgrades(self, only_profiles=None, propose_deferrable=True):
        profiles = self.gatherer.get_profiles(proposed_only=True,
                                              propose_deferrable=propose_deferrable)
        if only_profiles:
            profiles = filter(lambda profile: profile['id'] in only_profiles, profiles)
        if not profiles:
            return []
        return reduce(list.__add__, map(itemgetter('upgrades'), profiles))

    def _validate_upgrade_ids(self, *api_ids):
        self.gatherer.get_upgrades_by_api_ids(*api_ids)

    def _validate_profile_ids(self, *profiles):
        for profile in profiles:
            # Note: profileExists can handle ids with or without 'profile-' at
            # the start.
            if not self.portal_setup.profileExists(profile):
                raise ProfileNotFound(profile)

    def _install_upgrades(self, *api_ids, **kwargs):
        propose_deferrable = kwargs.pop('propose_deferrable', True)
        executioner = IExecutioner(self.portal_setup)
        try:
            with ResponseLogger(self.request.RESPONSE, annotate_result=True):
                executioner.install_upgrades_by_api_ids(
                    *api_ids, propose_deferrable=propose_deferrable)
        except Exception, exc:
            raise AbortTransactionWithStreamedResponse(exc)

    def _install_profiles(self, *profile_ids, **options):
        executioner = IExecutioner(self.portal_setup)
        try:
            with ResponseLogger(self.request.RESPONSE, annotate_result=True):
                executioner.install_profiles_by_profile_ids(
                    *profile_ids, **options)
        except Exception, exc:
            raise AbortTransactionWithStreamedResponse(exc)

    def _require_up_to_date_plone_site(self):
        portal_migration = getToolByName(self.context, 'portal_migration')
        if portal_migration.needUpgrading():
            raise PloneSiteOutdated()



from datetime import datetime
from ftw.builder import Builder
from ftw.builder import create
from ftw.upgrade import UpgradeStep
from ftw.upgrade.executioner import Executioner
from ftw.upgrade.indexing import HAS_INDEXING
from ftw.upgrade.interfaces import IExecutioner
from ftw.upgrade.tests.base import UpgradeTestCase
from Products.CMFCore.utils import getToolByName
from unittest2 import skipIf
from zope.component import queryAdapter
from zope.interface.verify import verifyClass
import transaction


class TestExecutioner(UpgradeTestCase):

    def test_component_is_registered(self):
        setup_tool = getToolByName(self.layer['portal'], 'portal_setup')
        executioner = queryAdapter(setup_tool, IExecutioner)
        self.assertNotEqual(None, executioner)

    def test_implements_interface(self):
        verifyClass(IExecutioner, Executioner)

    def test_installs_upgrades(self):
        def upgrade(setup_context):
            portal = setup_context.portal_url.getPortalObject()
            portal.upgrade_step_executed = True

        self.package.with_profile(Builder('genericsetup profile')
                                   .with_upgrade(Builder('plone upgrade step')
                                                 .upgrading('1000', to='1002')
                                                 .calling(upgrade)))

        self.portal.upgrade_step_executed = False
        with self.package_created():
            self.install_profile('the.package:default', version='1000')
            self.assertFalse(self.portal.upgrade_step_executed)
            self.install_profile_upgrades('the.package:default')
            self.assertTrue(self.portal.upgrade_step_executed)

    def test_install_upgrades_by_api_ids(self):
        def upgrade(setup_context):
            portal = setup_context.portal_url.getPortalObject()
            portal.upgrade_step_executed = True

        self.package.with_profile(Builder('genericsetup profile')
                                   .with_upgrade(Builder('plone upgrade step')
                                                 .upgrading('1000', to='1002')
                                                 .calling(upgrade)))

        self.portal.upgrade_step_executed = False
        with self.package_created():
            self.install_profile('the.package:default', version='1000')
            self.assertFalse(self.portal.upgrade_step_executed)
            executioner = queryAdapter(self.portal_setup, IExecutioner)
            executioner.install_upgrades_by_api_ids('1002@the.package:default')
            self.assertTrue(self.portal.upgrade_step_executed)

    def test_transaction_note(self):
        self.package.with_profile(
            Builder('genericsetup profile')
            .with_upgrade(Builder('plone upgrade step')
                          .upgrading('1000', to='1001')
                          .titled(u'Register "foo" utility'))
            .with_upgrade(Builder('plone upgrade step')
                          .upgrading('1001', to='1002')
                          .titled(u'Update email address'))
            .with_upgrade(Builder('plone upgrade step')
                          .upgrading('1002', to='1003')
                          .titled(u'Update email from name')))

        with self.package_created():
            self.install_profile('the.package:default', version='1000')
            self.install_profile_upgrades('the.package:default')
            self.assertMultiLineEqual(
                u'the.package:default -> 1001 (Register "foo" utility)\n'
                u'the.package:default -> 1002 (Update email address)\n'
                u'the.package:default -> 1003 (Update email from name)',
                transaction.get().description)

    def test_after_commit_hook(self):
        self.package.with_profile(
            Builder('genericsetup profile')
            .with_upgrade(Builder('plone upgrade step')
                          .upgrading('1000', to='1001')
                          .titled(u'Register "foo" utility')))

        with self.package_created():
            self.install_profile('the.package:default', version='1000')
            self.install_profile_upgrades('the.package:default')

            hooks = list(transaction.get().getAfterCommitHooks())
            hook_funcs = [h[0] for h in hooks]

            self.assertIn(
                'notification_hook',
                [f.func_name for f in hook_funcs],
                'Our notification_hook should be registered')

            transaction.commit()
            self.assertEquals(
                [],
                list(transaction.get().getAfterCommitHooks()),
                'Hook registrations should not persist across transactions')

    def test_resources_are_recooked_after_installing_upgrades(self):
        self.package.with_profile(
            Builder('genericsetup profile')
            .with_upgrade(Builder('plone upgrade step').upgrading('1000', to='1001')))
        with self.package_created():
            self.install_profile('the.package:default', version='1000')
            with self.assert_resources_recooked():
                self.install_profile_upgrades('the.package:default')

    def test_updates_quickinstaller_version(self):
        quickinstaller = getToolByName(self.portal, 'portal_quickinstaller')

        self.package.with_profile(Builder('genericsetup profile')
                                  .with_upgrade(Builder('plone upgrade step')
                                                .upgrading('1000', to='1001')))
        self.package.with_version('1.1')

        with self.package_created():
            self.install_profile('the.package:default', version='1000')
            quickinstaller.get('the.package').installedversion = '1.0'
            self.assertEquals('1.0', quickinstaller.get('the.package').getInstalledVersion())
            self.install_profile_upgrades('the.package:default')
            self.assertEquals('1.1', quickinstaller.get('the.package').getInstalledVersion())

    def test_install_profiles_by_profile_ids(self):
        self.package.with_profile(Builder('genericsetup profile')
                                  .with_upgrade(Builder('plone upgrade step')
                                                .upgrading('1000', to='1001')))

        self.setup_logging()
        profile_id = 'the.package:default'
        with self.package_created():
            executioner = queryAdapter(self.portal_setup, IExecutioner)
            executioner.install_profiles_by_profile_ids(profile_id)
            self.assertEquals(
                ['Installing profile the.package:default.',
                 'Done installing profile the.package:default.'],
                self.get_log())

            self.purge_log()
            executioner.install_profiles_by_profile_ids(profile_id)
            self.assertEquals(
                ['Ignoring already installed profile the.package:default.'],
                self.get_log())

            self.purge_log()
            executioner.install_profiles_by_profile_ids(profile_id,
                                                        force_reinstall=True)
            self.assertEquals(
                ['Installing profile the.package:default.',
                 'Done installing profile the.package:default.'],
                self.get_log())

    def test_do_not_decrease_version_when_only_installing_orphan_steps(self):
        self.package.with_profile(
            Builder('genericsetup profile')
            .with_upgrade(Builder('ftw upgrade step')
                          .to(datetime(2011, 11, 11, 11, 11)))
            .with_upgrade(Builder('ftw upgrade step')
                          .to(datetime(2012, 12, 12, 12, 12))))

        with self.package_created():
            self.install_profile('the.package:default', '20121212121200')
            self.clear_recorded_upgrades('the.package:default')
            self.record_installed_upgrades('the.package:default', '20121212121200')
            self.assert_gathered_upgrades({
                'the.package:default': {
                    'done': ['20121212121200'],
                    'proposed': ['20111111111100'],
                    'orphan': ['20111111111100']}})
            self.assertEquals(
                (u'20121212121200',),
                self.portal_setup.getLastVersionForProfile('the.package:default'))

            executioner = queryAdapter(self.portal_setup, IExecutioner)
            executioner.install_upgrades_by_api_ids(
                '20111111111100@the.package:default')
            self.assertEquals(
                (u'20121212121200',),
                self.portal_setup.getLastVersionForProfile('the.package:default'))
            self.assert_gathered_upgrades({
                'the.package:default': {
                    'done': ['20111111111100', '20121212121200'],
                    'proposed': [],
                    'orphan': []}})

    @skipIf(not HAS_INDEXING,
            'Tests must only run when indexing is available')
    def test_logs_indexing_progress_of_final_reindex(self):
        self.grant('Manager')
        create(Builder('folder'))
        create(Builder('folder'))

        class TriggerReindex(UpgradeStep):
            def __call__(self):
                catalog = self.getToolByName("portal_catalog")
                for brain in catalog(portal_type="Folder"):
                    brain.getObject().reindexObject()

        self.package.with_profile(
            Builder('genericsetup profile')
            .with_upgrade(Builder('ftw upgrade step')
                          .to(datetime(2013, 1, 1))
                          .calling(TriggerReindex)))

        with self.package_created():
            self.install_profile('the.package:default', version='1000')
            self.setup_logging()

            self.install_profile_upgrades('the.package:default')
            self.assertEquals(
                '1 of 2 (50%): Processing indexing queue',
                self.get_log()[-1])

    def test_regression_switching_versioning_system(self):
        # test_do_not_decrease_version_when_only_installing_orphan_steps
        # caused an issue when switching to upgrade step directories,
        # since the versioning system changed and '4' > '2016' but 4 < 2016.

        self.package.with_profile(
            Builder('genericsetup profile')
            .with_upgrade(Builder('plone upgrade step')
                          .upgrading('4000', to='4002'))
            .with_upgrade(Builder('ftw upgrade step')
                          .to(datetime(2011, 11, 11, 11, 11))))

        with self.package_created():
            self.install_profile('the.package:default', '4002')
            self.clear_recorded_upgrades('the.package:default')
            self.assert_gathered_upgrades({
                'the.package:default': {
                    'done': ['4002'],
                    'proposed': ['20111111111100'],
                    'orphan': []}})
            self.assertEquals(
                (u'4002',),
                self.portal_setup.getLastVersionForProfile('the.package:default'))

            executioner = queryAdapter(self.portal_setup, IExecutioner)
            executioner.install_upgrades_by_api_ids(
                '20111111111100@the.package:default')
            self.assertEquals(
                (u'20111111111100',),
                self.portal_setup.getLastVersionForProfile('the.package:default'))
            self.assert_gathered_upgrades({
                'the.package:default': {
                    'done': ['4002', '20111111111100'],
                    'proposed': [],
                    'orphan': []}})



from datetime import datetime
from ftw.builder import Builder
from ftw.builder import create
from ftw.upgrade.interfaces import IRecordableHandler
from ftw.upgrade.interfaces import IUpgradeStep
from ftw.upgrade.tests.base import UpgradeTestCase
from ftw.upgrade.utils import get_sorted_profile_ids
from Products.CMFCore.utils import getToolByName
from Products.CMFPlone.interfaces import IMigratingPloneSiteRoot
from Products.GenericSetup.interfaces import EXTENSION
from Products.GenericSetup.upgrade import listUpgradeSteps
from zope.configuration.config import ConfigurationExecutionError
from zope.interface import implementer
from zope.interface import Interface
from zope.interface import providedBy


class IFoo(Interface):
    """Dummy interface.
    """


class TestDirectoryMetaDirective(UpgradeTestCase):

    def setUp(self):
        super(TestDirectoryMetaDirective, self).setUp()
        self.profile = Builder('genericsetup profile')
        self.package.with_profile(self.profile)

    def test_upgrade_steps_are_registered(self):
        self.profile.with_upgrade(Builder('ftw upgrade step')
                                  .to(datetime(2011, 1, 1, 8))
                                  .named('add_action'))
        self.profile.with_upgrade(Builder('ftw upgrade step')
                                  .to(datetime(2011, 2, 2, 8))
                                  .named('remove_action'))

        with self.package_created():
            self.assert_upgrades([
                    {'source': ('10000000000000',),
                     'dest': ('20110101080000',),
                     'title': u'Add action.'},

                    {'source': ('20110101080000',),
                     'dest': ('20110202080000',),
                     'title': u'Remove action.'}])

    def test_first_source_version_is_last_regulare_upgrade_step(self):
        self.profile.with_upgrade(Builder('plone upgrade step')
                                  .upgrading('1000', to='1001')
                                  .titled(u'Register foo utility.'))
        self.profile.with_upgrade(Builder('ftw upgrade step')
                                  .to(datetime(2011, 1, 1, 8))
                                  .named('add_action'))

        with self.package_created():
            self.assert_upgrades([
                    {'source': ('1000',),
                     'dest': ('1001',),
                     'title': u'Register foo utility.'},

                    {'source': ('1001',),
                     'dest': ('20110101080000',),
                     'title': u'Add action.'}])

    def test_registers_migration_generic_setup_profile_foreach_step(self):
        self.profile.with_upgrade(Builder('ftw upgrade step')
                                  .to(datetime(2011, 1, 1, 8))
                                  .named('add_an_action'))

        with self.package_created() as package:
            upgrade_path = package.package_path.joinpath(
                'upgrades', '20110101080000_add_an_action')
            self.assert_profile(
                {'id': 'the.package.upgrades:default-upgrade-20110101080000',
                 'title': 'Upgrade the.package:default ' + \
                     'to 20110101080000: Add an action.',
                 'description': '',
                 'path': upgrade_path,
                 'product': 'the.package.upgrades',
                 'type': EXTENSION,
                 'for': IMigratingPloneSiteRoot})

    def test_package_modules_is_not_corrupted(self):
        # Regression: when the upgrade-step:directory directive is used from
        # the package-directory with a relative path (directory="upgrades"),
        # it corrupted the sys.modules entry of the package.

        package_builder = (
            Builder('python package')
            .named('other.package')
            .at_path(self.layer['temp_directory'])
            .with_file('__init__.py', 'PACKAGE = "package root"')

            .with_directory('profiles/default')
            .with_zcml_node('genericsetup:registerProfile',
                            name='default',
                            title='other.package:default',
                            directory='profiles/default',
                            provides='Products.GenericSetup.interfaces.EXTENSION')

            .with_directory('upgrades')
            .with_file('upgrades/__init__.py', 'PACKAGE = "upgrades package"')
            .with_zcml_include('ftw.upgrade', file='meta.zcml')
            .with_zcml_node('upgrade-step:directory',
                            profile='other.package:default',
                            directory='upgrades'))

        with create(package_builder).zcml_loaded(self.layer['configurationContext']):
            import other.package
            self.assertEquals('package root', other.package.PACKAGE)

    def test_profile_must_be_registed_before_registering_upgrade_directory(self):
        package_builder = (Builder('python package')
                           .named('other.package')
                           .at_path(self.layer['temp_directory'])
                           .with_zcml_include('ftw.upgrade', file='meta.zcml')
                           .with_zcml_node('upgrade-step:directory',
                                           profile='other.package:default',
                                           directory='.'))

        with create(package_builder) as package:
            with self.assertRaises(ConfigurationExecutionError) as cm:
                package.load_zcml(self.layer['configurationContext'])

        self.assertEqual(
            "<class 'ftw.upgrade.exceptions.UpgradeStepConfigurationError'>: "
            'The profile "other.package:default" needs to be registered'
            ' before registering its upgrade step directory.',
            str(cm.exception).splitlines()[0])

    def test_profile_version_is_set_to_latest_profile_version(self):
        self.profile.with_upgrade(Builder('ftw upgrade step').to(datetime(2011, 1, 1, 8)))
        self.profile.with_upgrade(Builder('ftw upgrade step').to(datetime(2011, 2, 2, 8)))

        with self.package_created() as package:
            profile_path = package.package_path.joinpath('profiles', 'default')
            self.assert_profile(
                {'id': u'the.package:default',
                 'title': u'the.package',
                 'description': u'',
                 'ftw.upgrade:dependencies': None,
                 'path': profile_path,
                 'version': '20110202080000',
                 'product': 'the.package',
                 'type': EXTENSION,
                 'for': None})

    def test_profile_version_is_set_to_latest_old_school_profile_version(self):
        self.profile.with_upgrade(Builder('plone upgrade step')
                                  .upgrading('1000', to='1001')
                                  .titled(u'Register foo utility.'))
        self.profile.with_upgrade(Builder('ftw upgrade step')
                                  .to(datetime(2011, 2, 2, 8)))

        package = create(self.package)
        # Remove upgrade-step directory upgrade in order to have the
        # manually created upgrade step as last version
        # but still declaring an upgrade-step:directory:
        package.package_path.joinpath(
            'upgrades', '20110202080000_upgrade').rmtree()

        profile_path = package.package_path.joinpath('profiles', 'default')
        self.assertNotIn('<version',
                         profile_path.joinpath('metadata.xml').text())

        with package.zcml_loaded(self.layer['configurationContext']):
            from ftw.upgrade.directory.zcml import find_start_version
            find_start_version(u'the.package:default')
            self.assert_profile(
                {'id': u'the.package:default',
                 'title': u'the.package',
                 'description': u'',
                 'ftw.upgrade:dependencies': None,
                 'path': str(profile_path),
                 'version': '1001',
                 'product': 'the.package',
                 'type': EXTENSION,
                 'for': None})

    def test_version_set_to_default_when_no_upgrades_defined(self):
        upgrades = self.package.package.get_subpackage('upgrades')
        upgrades.with_zcml_include('ftw.upgrade', file='meta.zcml')
        upgrades.with_zcml_node('upgrade-step:directory',
                                profile='the.package:default',
                                directory='.')

        with self.package_created() as package:
            profile_path = package.package_path.joinpath('profiles', 'default')
            self.assert_profile(
                {'id': u'the.package:default',
                 'title': u'the.package',
                 'description': u'',
                 'ftw.upgrade:dependencies': None,
                 'path': profile_path,
                 'version': u'10000000000000',
                 'product': 'the.package',
                 'type': EXTENSION,
                 'for': None})

    def test_profile_must_not_have_a_metadata_version_defined(self):
        self.profile.with_fs_version('1000')
        self.profile.with_upgrade(Builder('ftw upgrade step').to(datetime(2011, 1, 1, 8)))

        with create(self.package) as package:
            with self.assertRaises(ConfigurationExecutionError) as cm:
                package.load_zcml(self.layer['configurationContext'])

        self.assertEqual(
            "<class 'ftw.upgrade.exceptions.UpgradeStepConfigurationError'>: "
            'Registering an upgrades directory for "the.package:default" requires'
            ' this profile to not define a version in its metadata.xml.'
            ' The version is automatically set to the latest upgrade.',
            str(cm.exception).splitlines()[0])

    def test_declaring_upgrades_dependency(self):
        self.package.with_profile(
            Builder('genericsetup profile')
            .named('bar')
            .with_upgrade(
                Builder('ftw upgrade step')
                .to(datetime(2010, 1, 1, 1, 1))
                .with_zcml_directory_options(
                    soft_dependencies="the.package:baz")))

        self.package.with_profile(
            Builder('genericsetup profile')
            .named('foo')
            .with_upgrade(
                Builder('ftw upgrade step')
                .to(datetime(2010, 1, 1, 1, 1))
                .with_zcml_directory_options(
                    soft_dependencies="the.package:bar the.package:baz")))

        self.package.with_profile(
            Builder('genericsetup profile')
            .named('baz')
            .with_upgrade(
                Builder('ftw upgrade step')
                .to(datetime(2010, 1, 1, 1, 1))
                .with_zcml_directory_options(
                    soft_dependencies="the.package:default")))

        with self.package_created() as package:
            self.assert_profile(
                {'id': u'the.package:foo',
                 'title': u'the.package',
                 'description': u'',
                 'ftw.upgrade:dependencies': [u'the.package:bar',
                                              u'the.package:baz'],
                 'path': package.package_path.joinpath('profiles', 'foo'),
                 'version': u'20100101010100',
                 'product': 'the.package',
                 'type': EXTENSION,
                 'for': None})

            portal_setup = getToolByName(self.portal, 'portal_setup')
            self.assertEquals(
                [u'the.package:default',
                 u'the.package:baz',
                 u'the.package:bar',
                 u'the.package:foo'],
                filter(lambda profile_id: profile_id.startswith('the.package:'),
                       get_sorted_profile_ids(portal_setup)))

    def test_handler_step_provides_interfaces_implemented_by_upgrade_step_class(self):
        code = '\n'.join((
            'from ftw.upgrade import UpgradeStep',
            'from ftw.upgrade.tests.test_directory_meta_directive import IFoo',
            'from zope.interface import implementer',
            '',
            '@implementer(IFoo)',
            'class Foo(UpgradeStep):',
            '    pass'))

        self.profile.with_upgrade(Builder('ftw upgrade step')
                                  .to(datetime(2011, 1, 1, 1))
                                  .with_code(code))

        with self.package_created():
            portal_setup = getToolByName(self.portal, 'portal_setup')
            steps = listUpgradeSteps(portal_setup, 'the.package:default', '10000000000000')
            self.assertEquals(1, len(steps))
            self.assertItemsEqual(
                (IRecordableHandler,
                 IUpgradeStep,
                 IFoo),
                tuple(providedBy(steps[0]['step'].handler)))
            self.assertTrue(steps[0]['step'].handler.handler)

    def assert_upgrades(self, expected):
        upgrades = self.portal_setup.listUpgrades('the.package:default')
        got = [dict((key, value) for (key, value) in step.items()
                    if key in ('source', 'dest', 'title'))
               for step in upgrades]
        self.maxDiff = None
        self.assertItemsEqual(expected, got)

    def assert_profile(self, expected):
        self.assertTrue(
            self.portal_setup.profileExists(expected['id']),
            'Profile "{0}" does not exist. Profiles: {1}'.format(
                expected['id'],
                [profile['id'] for profile in self.portal_setup.listProfileInfo()]))

        got = self.portal_setup.getProfileInfo(expected['id']).copy()

        # Ignore pre_handler and post_handler, only available in Plone >= 4.3.8
        got.pop('pre_handler', None)
        got.pop('post_handler', None)

        self.maxDiff = None
        self.assertDictEqual(expected, got)



from contextlib import contextmanager
from datetime import datetime
from ftw.builder import Builder
from ftw.builder import create
from ftw.upgrade.directory.scanner import Scanner
from ftw.upgrade.exceptions import UpgradeStepDefinitionError
from ftw.upgrade.tests.base import UpgradeTestCase


class TestDirectoryScanner(UpgradeTestCase):

    def setUp(self):
        super(TestDirectoryScanner, self).setUp()
        self.profile = Builder('genericsetup profile')
        self.package.with_profile(self.profile)

    def test_returns_chained_upgrade_infos(self):
        self.profile.with_upgrade(Builder('ftw upgrade step')
                                  .to(datetime(2011, 1, 1, 8))
                                  .named('add an action'))
        self.profile.with_upgrade(Builder('ftw upgrade step')
                                  .to(datetime(2011, 2, 2, 8))
                                  .named('update the action'))
        self.profile.with_upgrade(Builder('ftw upgrade step')
                                  .to(datetime(2011, 3, 3, 8))
                                  .named('remove the action'))

        with self.scanned() as upgrade_infos:
            map(lambda info: (info.__delitem__('path'),
                              info.__delitem__('callable')),
                upgrade_infos)

            self.maxDiff = None
            self.assertEqual(
                [{'source-version': None,
                  'target-version': '20110101080000',
                  'title': 'Add an action.'},

                 {'source-version': '20110101080000',
                  'target-version': '20110202080000',
                  'title': 'Update the action.'},

                 {'source-version': '20110202080000',
                  'target-version': '20110303080000',
                  'title': 'Remove the action.'}],

                upgrade_infos)

    def test_exception_raised_when_upgrade_has_no_code(self):
        self.profile.with_upgrade(Builder('ftw upgrade step')
                                  .to(datetime(2011, 1, 1, 8))
                                  .named('add action')
                                  .with_code(''))

        with create(self.package) as package:
            with self.assertRaises(UpgradeStepDefinitionError) as cm:
                self.scan(package)

        self.assertEqual(
            'The upgrade step 20110101080000_add_action has no upgrade class'
            ' in the upgrade.py module.',
            str(cm.exception))

    def test_exception_raised_when_multiple_upgrade_steps_detected(self):
        code = '\n'.join((
                'from ftw.upgrade import UpgradeStep',
                'class Foo(UpgradeStep): pass',
                'class Bar(UpgradeStep): pass'))

        self.profile.with_upgrade(Builder('ftw upgrade step')
                                  .to(datetime(2011, 1, 1, 8))
                                  .named('add action')
                                  .with_code(code))

        with create(self.package) as package:
            with self.assertRaises(UpgradeStepDefinitionError) as cm:
                self.scan(package)

        self.assertEqual(
            'The upgrade step 20110101080000_add_action has more than one upgrade'
            ' class in the upgrade.py module.',
            str(cm.exception))

    def test_does_not_fail_when_no_upgrades_present(self):
        self.package.with_zcml_include('ftw.upgrade', file='meta.zcml')
        self.package.with_zcml_node('upgrade-step:directory',
                                    profile='the.package:default',
                                    directory='.')

        with self.scanned() as upgrade_infos:
            self.assertEqual( [], upgrade_infos)

    @contextmanager
    def scanned(self):
        with create(self.package) as package:
            yield self.scan(package)

    def scan(self, package):
        upgrades = package.package_path.joinpath('upgrades')
        return Scanner('the.package.upgrades', upgrades).scan()



from ftw.upgrade.tests.base import CommandTestCase
from datetime import datetime
from ftw.builder import Builder
from ftw.builder import create


class TestTouchCommand(CommandTestCase):

    def setUp(self):
        self.package_builder = (Builder('python package')
                                .named('the.package')
                                .at_path(self.layer.sample_buildout))
        self.package = None

    def test_touching_an_upgrade_step_renews_the_timestamp(self):
        self.package = create(
            self.package_builder
            .with_profile(Builder('genericsetup profile')
                          .with_upgrade(Builder('ftw upgrade step')
                                        .to(datetime(2011, 1, 1))
                                        .named('AddAction'))))

        path = self.package.package_path.joinpath('upgrades', '20110101000000_add_action')
        self.assertTrue(
            path.exists(),
            'Expected path to exist: {0}'.format(path))
        self.upgrade_script('touch {0}'.format(path))
        self.assertFalse(path.exists(),
                         'Expected path to no longer exist: {0}'.format(path))

        new_step_path, = path.dirname().dirs()
        self.assertRegexpMatches(new_step_path.name,
                                 r'^{0}\d{{10}}_add_action'.format(datetime.now().year))

    def test_moving_after_another(self):
        self.package = create(
            self.package_builder
            .with_profile(Builder('genericsetup profile')
                          .with_upgrade(Builder('ftw upgrade step')
                                        .to(datetime(2011, 11, 5))
                                        .named('add_action'))
                          .with_upgrade(Builder('ftw upgrade step')
                                        .to(datetime(2011, 11, 25))
                                        .named('remove_action'))
                          .with_upgrade(Builder('ftw upgrade step')
                                        .to(datetime(2022, 12, 12))
                                        .named('update_action'))))

        self.upgrade_script('touch {update_action} --after {add_action}'.format(
                **self.upgrades()))
        self.assert_upgrades('20111105000000_add_action',
                             '20111115000000_update_action',
                             '20111125000000_remove_action')

    def test_moving_explicitly_to_the_end(self):
        self.package = create(
            self.package_builder
            .with_profile(Builder('genericsetup profile')
                          .with_upgrade(Builder('ftw upgrade step')
                                        .to(datetime(2022, 12, 12, 0, 0, 0))
                                        .named('remove_action'))
                          .with_upgrade(Builder('ftw upgrade step')
                                        .to(datetime(2011, 11, 5, 0, 0, 55))
                                        .named('add_action'))))

        self.upgrade_script('touch {remove_action} --after {add_action}'.format(
                **self.upgrades()))

        self.assert_upgrades('20111105000055_add_action',
                             '20111106000055_remove_action')

    def test_moving_before_another(self):
        self.package = create(
            self.package_builder
            .with_profile(Builder('genericsetup profile')
                          .with_upgrade(Builder('ftw upgrade step')
                                        .to(datetime(2011, 11, 5))
                                        .named('add_action'))
                          .with_upgrade(Builder('ftw upgrade step')
                                        .to(datetime(2011, 11, 25))
                                        .named('remove_action'))
                          .with_upgrade(Builder('ftw upgrade step')
                                        .to(datetime(2022, 12, 12))
                                        .named('update_action'))))

        self.upgrade_script('touch {update_action} --before {remove_action}'.format(
                **self.upgrades()))

        self.assert_upgrades('20111105000000_add_action',
                             '20111115000000_update_action',
                             '20111125000000_remove_action')

    def test_moving_explicitly_to_the_beginning(self):
        self.package = create(
            self.package_builder
            .with_profile(Builder('genericsetup profile')
                          .with_upgrade(Builder('ftw upgrade step')
                                        .to(datetime(2011, 11, 6, 0, 0, 55))
                                        .named('remove_action'))
                          .with_upgrade(Builder('ftw upgrade step')
                                        .to(datetime(2022, 12, 12, 12, 12, 12))
                                        .named('add_action'))))

        self.upgrade_script('touch {add_action} --before {remove_action}'.format(
                **self.upgrades()))

        self.assert_upgrades('20111105000055_add_action',
                             '20111106000055_remove_action')

    def test_moving_before_and_after_at_same_time_is_not_allowed(self):
        self.package = create(
            self.package_builder
            .with_profile(Builder('genericsetup profile')
                          .with_upgrade(Builder('ftw upgrade step')
                                        .to(datetime(2011, 1, 1))
                                        .named('add_action'))
                          .with_upgrade(Builder('ftw upgrade step')
                                        .to(datetime(2011, 1, 2))
                                        .named('remove_action'))
                          .with_upgrade(Builder('ftw upgrade step')
                                        .to(datetime(2022, 1, 3))
                                        .named('update action'))))

        cmd = 'touch {add_action} --before {remove_action} --after {update_action}'.format(
            **self.upgrades())
        exitcode, output = self.upgrade_script(cmd, assert_exitcode=False)
        self.assertEqual(2, exitcode, 'command should fail because --after and --before can'
                         ' not be used at the same time.')
        self.assertIn('error: argument --after/-a: not allowed with argument --before/-b',
                      output)

    def test_moving_upgrade_step_should_not_import_upgrade_step(self):
        """The bin/upgrade path does usually only include ftw.upgrade (and dependencies)
        but not the path of the package in development or its dependencies.
        Therefore the bin/upgrade command should not import any code.
        """

        code = 'raise AssertionError("Upgrade code should not be imported..")'
        self.package = create(
            self.package_builder
            .with_profile(Builder('genericsetup profile')
                          .with_upgrade(Builder('ftw upgrade step')
                                        .to(datetime(2022, 12, 12, 0, 0, 0))
                                        .named('remove_action')
                                        .with_code(code))
                          .with_upgrade(Builder('ftw upgrade step')
                                        .to(datetime(2011, 11, 5, 0, 0, 55))
                                        .named('add_action')
                                        .with_code(code))))

        self.upgrade_script('touch {remove_action} --after {add_action}'.format(
                **self.upgrades()))

        self.assert_upgrades('20111105000055_add_action',
                             '20111106000055_remove_action')

    def upgrades(self):
        upgrades_dir = self.package.package_path.joinpath('upgrades')
        self.assertTrue(upgrades_dir.isdir(),
                        '"upgrades" directory is missing at {0}'.format(upgrades_dir))
        return dict([(path.name.split('_', 1)[1], path) for path in upgrades_dir.dirs()])

    def assert_upgrades(self, *expected):
        upgrades_dir = self.package.package_path.joinpath('upgrades')
        self.assertTrue(upgrades_dir.isdir(),
                        '"upgrades" directory is missing at {0}'.format(upgrades_dir))
        expected = set(expected)
        got = set([str(path.name) for path in upgrades_dir.dirs()])
        self.assertEqual(expected, got, 'Unexpected upgrades.')



from contextlib import closing
from ftw.upgrade.command.jsonapi import add_requestor_authentication_argument
from ftw.upgrade.command.jsonapi import add_site_path_argument
from ftw.upgrade.command.jsonapi import error_handling
from ftw.upgrade.command.jsonapi import with_api_requestor
from ftw.upgrade.command.terminal import TERMINAL
import sys


DOCS = """
{t.bold}DESCRIPTION:{t.normal}
    Upgrade Plone Site. \
This is what you would manually do in the @@plone-upgrade view.

{t.bold}EXAMPLES:{t.normal}
[quote]
$ bin/upgrade plone_upgrade --site Plone
[/quote]

""".format(t=TERMINAL).strip()


def setup_argparser(commands):
    command = commands.add_parser(
        'plone_upgrade',
        help='Upgrade Plone Site.',
        description=DOCS)
    command.set_defaults(func=plone_upgrade_command)
    add_requestor_authentication_argument(command)
    add_site_path_argument(command)


# expected output
expected = (
    u'Plone Site was already up to date.',
    u'Plone Site has been updated.')


@with_api_requestor
@error_handling
def plone_upgrade_command(args, requestor):
    action = 'plone_upgrade'
    params = ()

    with closing(requestor.POST(action, params=params,
                                stream=True)) as response:
        for line in response.iter_lines(chunk_size=30):
            if isinstance(line, unicode):
                line = line.encode('utf-8')

            print line

    line = line.strip()
    if not any([x in line for x in expected]):
        sys.exit(3)



from DateTime import DateTime
from Products.CMFCore.utils import getToolByName
from ftw.upgrade import ProgressLogger
from ftw.upgrade.helpers import update_security_for
from ftw.upgrade.utils import SavepointIterator
from ftw.upgrade.utils import SizedGenerator
from zope.component.hooks import getSite
import logging

LOG = logging.getLogger('ftw.upgrade.WorkflowChainUpdater')


class WorkflowChainUpdater(object):

    def __init__(self, objects, review_state_mapping, update_security=True,
                 migrate_workflow_history=True, transition_mapping=None):
        self.objects = tuple(objects)
        self.review_state_mapping = review_state_mapping
        self.update_security = update_security
        self.migrate_workflow_history = migrate_workflow_history
        self.transition_mapping = transition_mapping or {}
        self.started = False
        self.wfs_and_states_before = None

    def __enter__(self):
        assert not self.started, 'WorkflowChainUpdater was already started.'
        self._started = True

        self.wfs_and_states_before = self.get_workflows_and_states(
            self.get_objects())

        return self

    def __exit__(self, exc_type, exc_value, traceback):
        if exc_type:
            return None

        self.update_workflow_states_with_mapping()

    def get_objects(self):
        return self.objects

    def get_workflows_and_states(self, objects):
        title = 'Get workflows and review states'

        wftool = None
        result = {}
        for obj in ProgressLogger(title, objects):
            if wftool is None:
                wftool = getToolByName(obj, 'portal_workflow')

            path = '/'.join(obj.getPhysicalPath())
            workflow = self._get_workflow_id_for(obj, wftool)
            if workflow:
                review_state = wftool.getInfoFor(obj, 'review_state')
            else:
                review_state = None

            result[path] = {'workflow': workflow,
                            'review_state': review_state}

        return result

    def update_workflow_states_with_mapping(self):
        status_before_activation = self.wfs_and_states_before
        status_after_activation = self.get_workflows_and_states(
            self.get_objects())

        LOG.info('Changing workflow states of objects which were'
                 ' reset to the initial state according to mapping.')

        portal = getSite()
        wf_tool = getToolByName(portal, 'portal_workflow')
        origin_workflows = zip(*self.review_state_mapping.keys())[0]

        title = 'Change workflow states'
        for path in ProgressLogger(title, status_before_activation):
            wf_before = status_before_activation[path].get('workflow')
            review_state_before = status_before_activation[path].get(
                'review_state')
            wf_after = status_after_activation[path].get('workflow')

            if wf_before not in origin_workflows:
                # This object has not a workflow which is in the
                # mapping, thus no migration is needed.
                continue

            mapping = self.review_state_mapping.get(
                (wf_before, wf_after), {})
            new_review_state = mapping.get(review_state_before)
            if not new_review_state:
                LOG.warn(
                    'Mapping not defined for old state %s when changing'
                    ' workflow from %s to %s.' % (
                        review_state_before, wf_before, wf_after))
                continue

            obj = portal.unrestrictedTraverse(path)

            if self.migrate_workflow_history:
                self._migrate_workflow_history(obj, wf_before, wf_after)
            else:
                wf_tool.setStatusOf(wf_after, obj, {
                    'review_state': new_review_state,
                    'action': 'systemupdate',
                    'actor': 'system',
                    'comments': '',
                    'time': DateTime()})

            if self.update_security:
                update_security_for(obj, reindex_security=True)
                obj.reindexObject(idxs=['review_state'])

    def _get_workflow_id_for(self, context, wftool):
        workflows = wftool.getWorkflowsFor(context)
        assert len(workflows) in (0, 1), \
            'Only one workflow per object supported. %s' % str(context)

        if len(workflows) == 0:
            return None

        else:
            return workflows[0].id

    def _migrate_workflow_history(self, context, old_wf, new_wf):
        wfhistory = getattr(context, 'workflow_history', None)
        if wfhistory is None or old_wf not in wfhistory:
            return

        def _migrate_action(entry):
            action = entry.get('action', None)
            if action:
                actionmapping = self.transition_mapping.get(
                    (old_wf, new_wf), {})
                if action in actionmapping:
                    entry['action'] = actionmapping[action]

            state = entry.get('review_state', None)
            if state:
                statemapping = self.review_state_mapping.get((old_wf, new_wf),
                                                             {})
                if state in statemapping:
                    entry['review_state'] = statemapping[state]

        def _migrate_entry(entry):
            entry = entry.copy()
            _migrate_action(entry)
            return entry

        wfhistory[new_wf] = map(_migrate_entry, wfhistory[old_wf])


class WorkflowSecurityUpdater(object):

    def update(self, changed_workflows, reindex_security=True, savepoints=None):
        types = self.get_suspected_types(changed_workflows)
        objects = SavepointIterator.build(self.lookup_objects(types), savepoints)
        for obj in objects:
            if self.obj_has_workflow(obj, changed_workflows):
                update_security_for(obj, reindex_security=reindex_security)

    def lookup_objects(self, types):
        portal = getSite()
        catalog = getToolByName(portal, 'portal_catalog')

        query = {'portal_type': types}
        brains = tuple(catalog.unrestrictedSearchResults(query))

        lookup = lambda brain: portal.unrestrictedTraverse(brain.getPath())
        generator = SizedGenerator((lookup(brain) for brain in brains),
                                   len(brains))
        return ProgressLogger('Update object security', generator)

    def get_suspected_types(self, changed_workflows):
        types = []
        ttool = getToolByName(getSite(), 'portal_types')

        for fti in ttool.objectValues():
            portal_type = fti.getId()
            if self.type_workflow_is_one_of(portal_type, changed_workflows):
                types.append(portal_type)

        return types

    def type_workflow_is_one_of(self, portal_type, workflows):
        wftool = getToolByName(getSite(), 'portal_workflow')
        default_chain = wftool.getChainForPortalType(portal_type)
        if set(default_chain) & set(workflows):
            return True

        try:
            pwftool = getToolByName(getSite(), 'portal_placeful_workflow')
        except AttributeError:
            return False

        for policy in pwftool.objectValues():
            chain = policy.getChainFor(portal_type) or []
            if set(chain) & set(workflows):
                return True

        return False

    def obj_has_workflow(self, obj, workflows):
        wftool = getToolByName(getSite(), 'portal_workflow')
        obj_workflow_names = map(lambda wf: wf.getId(),
                                 wftool.getWorkflowsFor(obj))
        return set(obj_workflow_names) & set(workflows)



from ftw.builder import builder_registry
from ftw.builder.utils import serialize_callable
from ftw.upgrade import UpgradeStep
from ftw.upgrade.directory import scaffold
from path import Path
import inflection
import os


class DeferrableUpgrade(UpgradeStep):

    deferrable = True

    def __call__(self):
        pass


class UpgradeStepBuilder(object):

    def __init__(self, session):
        self.session = session
        self.destination_version = None
        self.name = None
        self.package = None
        self.profile_builder = None
        self.named('Upgrade')
        self.code = None
        self.directories = []
        self.files = []
        self.zcml_directory_options = None

    def to(self, destination):
        if hasattr(destination, 'strftime'):
            self.destination_version = destination.strftime(scaffold.DATETIME_FORMAT)
        else:
            self.destination_version = destination
        return self

    def named(self, name):
        self.name = name
        return self

    def for_profile(self, profile_builder):
        self.profile_builder = profile_builder
        if self.profile_builder.fs_version is None:
            self.profile_builder.with_fs_version(False)
        return self

    def with_code(self, code_as_string):
        self.code = code_as_string
        return self

    def calling(self, callable_, *to_import):
        """Make the upgrade step execute the callable passed as argument.
        The callable will be serialized to a string.

        If the callable is a class, superclasses are automatically imported.
        Other globals are not imported and need to be passed to ``calling``
        as additional positional arguments.
        """

        source = serialize_callable(callable_, *to_import)
        return self.with_code(source)

    def as_deferrable(self):
        return self.calling(DeferrableUpgrade)

    def with_directory(self, relative_path):
        """Create a directory in the profile.
        """
        self.directories.append(relative_path)
        return self

    def with_zcml_directory_options(self, **options):
        """Set additional options in the upgrade-step:directory directive.
        """
        self.zcml_directory_options = options
        return self

    def with_file(self, relative_path, contents, makedirs=False):
        """Create a file within this package.
        """
        if makedirs and Path(relative_path).parent:
            self.with_directory(Path(relative_path).parent)

        self.files.append((relative_path, contents))
        return self

    def create(self):
        if self.destination_version is None:
            raise ValueError('A destination version is required.'
                             ' Use .to(datetime(...)).')
        self._set_package()
        self._declare_zcml_directory()
        name = self._create_upgrade()
        self._register_files_and_dirs_in_package_builder(name)

    def _set_package(self):
        self.package = self.profile_builder.package.get_subpackage('upgrades')
        if self.profile_builder.name != 'default':
            self.package = self.package.get_subpackage(self.profile_builder.name)

    def _declare_zcml_directory(self):
        zcml = self.package.get_configure_zcml()

        if getattr(zcml, '_upgrade_step_declarations', None) is None:
            zcml._upgrade_step_declarations = {}

        if zcml._upgrade_step_declarations.get(self.profile_builder.name):
            return

        zcml.include('ftw.upgrade', file='meta.zcml')
        zcml.with_node('upgrade-step:directory',
                       profile=self.profile_builder.profile_name,
                       directory='.',
                       **(self.zcml_directory_options or {}))
        zcml._upgrade_step_declarations[self.profile_builder.name] = True

    def _create_upgrade(self):
        name = self.name.replace(' ', '_').replace('\.$', '')
        step_name = '{0}_{1}'.format(self.destination_version,
                                     inflection.underscore(name))
        if self.code is None:
            self.code = scaffold.PYTHON_TEMPLATE.format(
                classname=inflection.camelize(name),
                docstring=inflection.humanize(
                    inflection.underscore(name)))

        self.package.with_file(
            os.path.join(step_name, '__init__.py'),
            '',
            makedirs=True)

        self.package.with_file(
            os.path.join(step_name, 'upgrade.py'),
            self.code)
        return step_name

    def _register_files_and_dirs_in_package_builder(self, step_name):
        for relative_path in self.directories:
            self.package.with_directory(os.path.join(step_name, relative_path))

        for path, contents in self.files:
            self.package.with_file(os.path.join(step_name, path), contents)


builder_registry.register('ftw upgrade step', UpgradeStepBuilder)



from ftw.builder.testing import BUILDER_LAYER
from ftw.builder.testing import functional_session_factory
from ftw.builder.testing import set_builder_session_factory
from ftw.testbrowser import browser
from ftw.testing.layer import COMPONENT_REGISTRY_ISOLATION
from ftw.testing.layer import ConsoleScriptLayer
from ftw.testing.layer import TEMP_DIRECTORY
from operator import itemgetter
from plone.app.testing import applyProfile
from plone.app.testing import FunctionalTesting
from plone.app.testing import PLONE_ZSERVER
from plone.app.testing import PloneSandboxLayer
from plone.app.testing import SITE_OWNER_NAME
from plone.registry.interfaces import IRegistry
from plone.testing import z2
from Products.CMFCore.utils import getToolByName
from Products.CMFPlone.utils import getFSVersionTuple
from Products.SiteAccess.VirtualHostMonster import manage_addVirtualHostMonster
from zope.component import getUtility
from zope.configuration import xmlconfig
import ftw.upgrade.tests.builders
import pkg_resources
import transaction


ftw.upgrade.tests.builders  # pyflakes


COMMAND_LAYER = ConsoleScriptLayer('ftw.upgrade',
                                   bases=(BUILDER_LAYER, ),
                                   name='ftw.upgrade:command')


class UpgradeLayer(PloneSandboxLayer):
    defaultBases = (COMPONENT_REGISTRY_ISOLATION,
                    BUILDER_LAYER,
                    TEMP_DIRECTORY)

    def setUpZope(self, app, configurationContext):
        import Products.CMFPlacefulWorkflow
        xmlconfig.file('configure.zcml', Products.CMFPlacefulWorkflow,
                       context=configurationContext)

        import ftw.upgrade
        xmlconfig.file('configure.zcml', ftw.upgrade,
                       context=configurationContext)

        z2.installProduct(app, 'Products.DateRecurringIndex')
        import plone.app.contenttypes
        xmlconfig.file('configure.zcml', plone.app.contenttypes,
                       context=configurationContext)

        z2.installProduct(app, 'Products.CMFPlacefulWorkflow')

        try:
            # Plone 4 with collective.indexing
            pkg_resources.get_distribution('collective.indexing')
        except pkg_resources.DistributionNotFound:
            pass
        else:
            import collective.indexing
            xmlconfig.file('configure.zcml', collective.indexing,
                           context=configurationContext)
            z2.installProduct(app, 'collective.indexing')

        manage_addVirtualHostMonster(app, 'virtual_hosting')

    def setUpPloneSite(self, portal):
        if getFSVersionTuple() > (5, ):
            applyProfile(portal, 'plone.app.contenttypes:default')

        applyProfile(
            portal, 'Products.CMFPlacefulWorkflow:CMFPlacefulWorkflow')
        applyProfile(portal, 'ftw.upgrade:default')

        self.fix_plone_app_jquery_version(portal)
        self.prevent_csrf_by_initializing_site_storages(portal)

        # Enable development mode so that resources are included in separately
        # in the HTML so that we can test for recooked resources.
        registry = getUtility(IRegistry)
        if 'plone.resources.development' in registry:
            registry['plone.resources.development'] = True

    def prevent_csrf_by_initializing_site_storages(self, portal):
        """Plone auto-protection results in confirmation pages
        when first hitting Plone with the browser.

        By hitting the site once we can initialize all standard Plone
        containers which cause writes on first hit.
        We do this with disabled protection.
        """
        try:
            from plone.protect import auto
        except ImportError:
            return

        transaction.commit()
        with browser(portal.aq_inner.aq_parent):
            crsrf_disabled_ori = auto.CSRF_DISABLED
            auto.CSRF_DISABLED = True
            try:
                browser.login(SITE_OWNER_NAME).open(
                    view='overview-controlpanel')
            finally:
                auto.CSRF_DISABLED = crsrf_disabled_ori
                transaction.begin()

    def fix_plone_app_jquery_version(self, portal):
        try:
            pkg_resources.get_distribution('plone.app.jquery')
        except pkg_resources.DistributionNotFound:
            return

        # The plone.app.jquery version shipped with Plone 4.2 has an outdated
        # metadata.xml version, resulting in proposed upgrades in a fresh
        # installation.
        # For consistent test result we fix that here.
        portal_setup = getToolByName(portal, 'portal_setup')
        profileid = 'plone.app.jquery:default'
        version = max(map(itemgetter('dest'),
                          portal_setup.listUpgrades(profileid, show_old=True)))
        portal_setup.setLastVersionForProfile(profileid, version)


UPGRADE_LAYER = UpgradeLayer()
UPGRADE_FUNCTIONAL_TESTING = FunctionalTesting(
    bases=(UPGRADE_LAYER,
           set_builder_session_factory(functional_session_factory)),
    name="ftw.upgrade:functional")

COMMAND_AND_UPGRADE_FUNCTIONAL_TESTING = FunctionalTesting(
    bases=(PLONE_ZSERVER,
           UPGRADE_LAYER,
           set_builder_session_factory(functional_session_factory),
           COMMAND_LAYER),
    name="ftw.upgrade:command_and_functional")


class IntIdUpgradeLayer(PloneSandboxLayer):

    defaultBases = (UPGRADE_LAYER, )

    def setUpZope(self, app, configurationContext):
        import plone.app.intid
        xmlconfig.file('configure.zcml', plone.app.intid,
                       context=configurationContext)

    def setUpPloneSite(self, portal):
        applyProfile(portal, 'plone.app.intid:default')


INTID_UPGRADE_LAYER = IntIdUpgradeLayer()
INTID_UPGRADE_FUNCTIONAL_TESTING = FunctionalTesting(
    bases=(INTID_UPGRADE_LAYER,
           set_builder_session_factory(functional_session_factory)),
    name="ftw.upgrade-intid:functional")



from ftw.upgrade.tests.base import CommandAndInstanceTestCase
from Products.CMFCore.utils import getToolByName
from Products.CMFPlone.utils import getFSVersionTuple
from unittest2 import skipIf
import transaction


class TestPloneUpgradeCommand(CommandAndInstanceTestCase):

    def setUp(self):
        super(TestPloneUpgradeCommand, self).setUp()
        self.write_zconf_with_test_instance()

    def test_help(self):
        self.upgrade_script('plone_upgrade --help')

    def test_plone_upgrade_already_uptodate(self):
        exitcode, output = self.upgrade_script(
            'plone_upgrade -s plone')
        self.assertEquals(0, exitcode)
        transaction.begin()  # sync transaction
        self.assertIn(u'Plone Site was already up to date.', output)

    @skipIf(getFSVersionTuple() > (5, ), 'The test does not work on Plone 5.')
    def test_upgrade_plone(self):
        from Products.CMFPlone.factory import _DEFAULT_PROFILE
        setup = getToolByName(self.portal, 'portal_setup')
        setup.setLastVersionForProfile(_DEFAULT_PROFILE, '4')
        transaction.commit()

        exitcode, output = self.upgrade_script(
            'plone_upgrade -s plone')
        self.assertEquals(0, exitcode)
        transaction.begin()  # sync transaction
        self.assertIn(u'Plone Site has been updated.', output)



from ftw.upgrade.interfaces import IClassMigratedEvent
from zope.component.interfaces import ObjectEvent
from zope.interface import implements


class ClassMigratedEvent(ObjectEvent):

    implements(IClassMigratedEvent)



from contextlib import contextmanager
from ftw.builder import Builder
from ftw.builder import create
from ftw.testbrowser import browser
from ftw.upgrade.directory import scaffold
from ftw.upgrade.indexing import processQueue
from ftw.upgrade.interfaces import IExecutioner
from ftw.upgrade.interfaces import IUpgradeInformationGatherer
from ftw.upgrade.interfaces import IUpgradeStepRecorder
from ftw.upgrade.testing import COMMAND_AND_UPGRADE_FUNCTIONAL_TESTING
from ftw.upgrade.testing import COMMAND_LAYER
from ftw.upgrade.testing import UPGRADE_FUNCTIONAL_TESTING
from ftw.upgrade.tests.helpers import verbose_logging
from operator import itemgetter
from path import Path
from plone.app.testing import setRoles
from plone.app.testing import SITE_OWNER_NAME
from plone.app.testing import TEST_USER_ID
from plone.app.testing import TEST_USER_PASSWORD
from Products.CMFCore.utils import getToolByName
from StringIO import StringIO
from unittest2 import TestCase
from zope.component import getMultiAdapter
from zope.component import queryAdapter
import json
import logging
import lxml.html
import os
import re
import transaction
import urllib


class UpgradeTestCase(TestCase):
    layer = UPGRADE_FUNCTIONAL_TESTING

    def setUp(self):
        self.package = (Builder('python package')
                        .at_path(self.directory)
                        .named('the.package'))
        self.portal = self.layer['portal']
        self.portal_setup = getToolByName(self.portal, 'portal_setup')
        self.portal_quickinstaller = getToolByName(self.portal, 'portal_quickinstaller')

    def tearDown(self):
        self.teardown_logging()

    def grant(self, *roles):
        setRoles(self.portal, TEST_USER_ID, list(roles))
        transaction.commit()

    @property
    def directory(self):
        return self.layer['temp_directory']

    @contextmanager
    def package_created(self):
        with create(self.package).zcml_loaded(self.layer['configurationContext']) as package:
            yield package

    def default_upgrade(self):
        return Builder('plone upgrade step').upgrading('1000', to='1001')

    def install_profile(self, profileid, version=None):
        self.portal_setup.runAllImportStepsFromProfile('profile-{0}'.format(profileid))
        if version is not None:
            self.portal_setup.setLastVersionForProfile(profileid, (unicode(version),))
        transaction.commit()

    def install_profile_upgrades(self, *profileids):
        gatherer = queryAdapter(self.portal_setup, IUpgradeInformationGatherer)
        upgrade_info = [(profile['id'], map(itemgetter('id'), profile['upgrades']))
                        for profile in gatherer.get_upgrades()
                        if profile['id'] in profileids]
        executioner = queryAdapter(self.portal_setup, IExecutioner)
        executioner.install(upgrade_info)

    def record_installed_upgrades(self, profile, *destinations):
        profile = re.sub('^profile-', '', profile)
        recorder = getMultiAdapter((self.portal, profile), IUpgradeStepRecorder)
        recorder.clear()
        map(recorder.mark_as_installed, destinations)
        transaction.commit()

    def clear_recorded_upgrades(self, profile):
        profile = re.sub('^profile-', '', profile)
        recorder = getMultiAdapter((self.portal, profile), IUpgradeStepRecorder)
        recorder.clear()
        transaction.commit()

    def assert_gathered_upgrades(self, expected, *args, **kwargs):
        gatherer = queryAdapter(self.portal_setup, IUpgradeInformationGatherer)
        result = gatherer.get_profiles(*args, **kwargs)
        got = {}
        for profile in result:
            if profile['id'] not in expected:
                continue

            got_profile = dict((key, []) for key in expected[profile['id']].keys())
            got[profile['id']] = got_profile

            for upgrade in profile['upgrades']:
                for key in got_profile.keys():
                    if upgrade[key]:
                        got_profile[key].append(upgrade['sdest'])

        self.maxDiff = None
        self.assertDictEqual(
            expected, got,
            'Unexpected gatherer result.\n\nPackages in result {0}:'.format(
                map(lambda profile: profile['id'], result)))

    def asset(self, filename):
        return Path(__file__).dirname().joinpath('assets', filename).text()

    @contextmanager
    def assert_resources_recooked(self):
        def get_resources():
            doc = lxml.html.fromstring(self.portal())
            return map(str.strip,
                       map(lxml.html.tostring,
                           doc.xpath('//link[@rel="stylesheet"][@href]'
                                     ' | //script[@src]')))

        resources = get_resources()
        yield
        self.assertNotEqual(resources, get_resources(),
                            'Resurces are not recooked.')

    def setup_logging(self):
        self.log = StringIO()
        self.loghandler = logging.StreamHandler(self.log)
        self.logger = logging.getLogger('ftw.upgrade')
        self.logger.setLevel(logging.DEBUG)
        self.logger.addHandler(self.loghandler)

    def teardown_logging(self):
        if getattr(self, 'log', None) is None:
            return

        self.logger.removeHandler(self.loghandler)
        self.log = None
        self.loghandler = None
        self.logger = None

    def get_log(self):
        return self.log.getvalue().splitlines()

    def purge_log(self):
        self.log.seek(0)
        self.log.truncate()


class CommandTestCase(TestCase):
    layer = COMMAND_LAYER

    def upgrade_script(self, args, assert_exitcode=True):
        command = ' '.join(('upgrade', args))
        exitcode, output = self.layer['execute_script'](
            command, assert_exitcode=assert_exitcode)

        output = (re.compile(r'/[^\n]*Terminal kind \'dumb\'[^\n]*\n', re.M)
                  .sub('', output))
        output = (re.compile(r'^[^\n]*_BINTERM_UNSUPPORTED[^\n]*\n', re.M)
                  .sub('', output))
        return exitcode, output


class WorkflowTestCase(TestCase):

    layer = UPGRADE_FUNCTIONAL_TESTING

    def setUp(self):
        self.portal = self.layer['portal']
        setRoles(self.portal, TEST_USER_ID, ['Manager'])

    def assertReviewStates(self, expected):
        wftool = getToolByName(self.portal, 'portal_workflow')

        got = {}
        for obj in expected.keys():
            review_state = wftool.getInfoFor(obj, 'review_state')
            got[obj] = review_state

        self.assertEquals(
            expected, got, 'Unexpected workflow states')

    def set_workflow_chain(self, for_type, to_workflow):
        wftool = getToolByName(self.portal, 'portal_workflow')
        wftool.setChainForPortalTypes((for_type,),
                                      (to_workflow,))

    def assertSecurityIsUpToDate(self):
        wftool = getToolByName(self.portal, 'portal_workflow')
        updated_objects = wftool.updateRoleMappings()
        self.assertEquals(
            0, updated_objects,
            'Expected all objects to have an up to date security, but'
            ' there were some which were not up to date.')

    def get_allowed_roles_and_users(self, for_object):
        processQueue()  # trigger async indexing
        catalog = getToolByName(self.portal, 'portal_catalog')
        path = '/'.join(for_object.getPhysicalPath())
        rid = catalog.getrid(path)
        index_data = catalog.getIndexDataForRID(rid)
        return index_data.get('allowedRolesAndUsers')

    def create_placeful_workflow_policy(self, named, with_workflows):
        placeful_wf_tool = getToolByName(
            self.portal, 'portal_placeful_workflow')

        placeful_wf_tool.manage_addWorkflowPolicy(named)
        policy = placeful_wf_tool.get(named)

        for portal_type, workflow in with_workflows.items():
            policy.setChain(portal_type, workflow)

        return policy

    def assert_permission_acquired(self, permission, obj, msg=None):
        not_acquired_permissions = self.get_not_acquired_permissions_of(obj)

        self.assertNotIn(
            permission, not_acquired_permissions,
            'Expected permission "%s" to be acquired on %s%s' % (
                permission, str(obj),
                msg and (' (%s)' % msg) or ''))

    def assert_permission_not_acquired(self, permission, obj, msg=None):
        not_acquired_permissions = self.get_not_acquired_permissions_of(obj)

        self.assertIn(
            permission, not_acquired_permissions,
            'Expected permission "%s" to NOT be acquired on %s%s' % (
                permission, str(obj),
                msg and (' (%s)' % msg) or ''))

    def get_not_acquired_permissions_of(self, obj):
        acquired_permissions = filter(
            lambda item: not item.get('acquire'),
            obj.permission_settings())

        return map(lambda item: item.get('name'), acquired_permissions)


class JsonApiTestCase(UpgradeTestCase):

    def assert_json_equal(self, expected, got, msg=None):
        expected = json.dumps(expected, sort_keys=True, indent=4)
        got = json.dumps(got, sort_keys=True, indent=4)
        self.maxDiff = None
        self.assertMultiLineEqual(expected, got, msg)

    def assert_json_contains_profile(self, expected_profileinfo, got, msg=None):
        profileid = expected_profileinfo['id']
        got_profiles = dict([(profile['id'], profile) for profile in got])
        self.assertIn(profileid, got_profiles,
                      'assert_json_contains_profile: expected profile not in JSON')
        self.assert_json_equal(expected_profileinfo, got_profiles[profileid], msg)

    def assert_json_contains(self, expected_element, got_elements):
        message = 'Could not find:\n\n{0}\n\nin list:\n\n{0}'.format(
            json.dumps(expected_element, sort_keys=True, indent=4),
            json.dumps(got_elements, sort_keys=True, indent=4))
        self.assertTrue(expected_element in got_elements, message)

    def is_installed(self, profileid, dest_time):
        recorder = getMultiAdapter((self.portal, profileid), IUpgradeStepRecorder)
        return recorder.is_installed(dest_time.strftime(scaffold.DATETIME_FORMAT))

    def api_request(self, method, action, data=(), authenticate=True,
                    context=None):
        if context is None:
            context = self.layer['portal']
        if authenticate:
            browser.login(SITE_OWNER_NAME)
        else:
            browser.logout()

        with verbose_logging():
            if method.lower() == 'get':
                browser.visit(context, view='upgrades-api/{0}?{1}'.format(
                        action,
                        urllib.urlencode(data)))

            elif method.lower() == 'post':
                if not data:
                    data = {'enforce': 'post'}
                browser.visit(context, view='upgrades-api/{0}'.format(action),
                              data=data)

            else:
                raise Exception('Unsupported request method {0}'.format(method))

    @contextmanager
    def expect_api_error(self, status=None, message=None, details=None):
        api_error_info = {}
        with browser.expect_http_error(code=status):
            yield api_error_info

        expected = {'result': 'ERROR'}
        if message is not None:
            expected['message'] = message
        if details is not None:
            expected['details'] = details

        got = dict(zip(['result', 'message', 'details'], browser.json))

        self.assertDictContainsSubset(
            expected,
            got,
            'Unexpected error response details.\n\n'
            'Expected:' +
            json.dumps(expected, sort_keys=True, indent=4) +
            '\nto be included in:\n' +
            json.dumps(got, sort_keys=True, indent=4))


class CommandAndInstanceTestCase(JsonApiTestCase, CommandTestCase):
    layer = COMMAND_AND_UPGRADE_FUNCTIONAL_TESTING

    def setUp(self):
        super(CommandAndInstanceTestCase, self).setUp()
        self.directory.joinpath('var').mkdir_p()
        os.environ['UPGRADE_AUTHENTICATION'] = ':'.join((SITE_OWNER_NAME,
                                                         TEST_USER_PASSWORD))

    def tearDown(self):
        if 'UPGRADE_AUTHENTICATION' in os.environ:
            del os.environ['UPGRADE_AUTHENTICATION']
        if 'UPGRADE_PUBLIC_URL' in os.environ:
            del os.environ['UPGRADE_PUBLIC_URL']

    @property
    def directory(self):
        return self.layer['root_path']

    def write_zconf(self, instance_name, port):
        etc1 = self.layer['root_path'].joinpath('parts', instance_name, 'etc')
        etc1.makedirs()
        etc1.joinpath('zope.conf').write_text(
            '\n'.join(('<http-server>',
                       '  address {0}'.format(port),
                       '</http-server>')))
        return etc1.dirname()

    def write_zconf_with_test_instance(self):
        test_instance_port = os.environ.get('ZSERVER_PORT', 55001)
        self.write_zconf('instance', test_instance_port)
